from .utils import (
    setup_logging,
    remove_ansi_escape_sequences,
    capture_subprocess_output,
)
from .output_logger import OutputLogger
from .job_logging_service import JobLoggingService, MissingJobLogException



from meltano.core.plugin import PluginInstall, PluginType
from meltano.core.plugin_invoker import PluginInvoker


class DbtInvoker(PluginInvoker):
    def __init__(self, project, plugin, **kwargs):
        kwargs["run_dir"] = project.root_dir("transform")
        return super().__init__(project, plugin, **kwargs)

    def Popen_options(self):
        options = super().Popen_options()
        options["cwd"] = str(self.project.root_dir("transform"))

        return options


class DbtPlugin(PluginInstall):
    __plugin_type__ = PluginType.TRANSFORMERS
    __invoker_cls__ = DbtInvoker

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)


class DbtTransformPlugin(PluginInstall):
    __plugin_type__ = PluginType.TRANSFORMS

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)



from datetime import date
from flask_security.utils import hash_password
from flask_security import AnonymousUser, SQLAlchemyUserDatastore

from meltano.api.models.security import db, User, Role, RolePermissions


users = SQLAlchemyUserDatastore(db, User, Role)


SEED_USERS = [
    {
        "username": "rob",
        "email": "rob@meltano.com",
        "password": "meltano",
        "confirmed_at": date(2000, 1, 1),
        "_roles": {"regular"},
    },
    {
        "username": "alice",
        "email": "alice@meltano.com",
        "password": "meltano",
        "confirmed_at": date(2000, 1, 1),
        "_roles": {"admin"},
    },
]


class FreeUser:
    """
    FreeUser is free to do everything and has no limits.

    Even though this class overrides `flask_security`'s AnonymousUser
    it doens't inherit from AnonymousUser to bypass some type check
    regarding the loading of the identity.
    """

    def has_role(*args):
        return True

    @property
    def username(self):
        return None

    @property
    def is_authenticated(self):
        return True

    @property
    def is_active(self):
        return True

    @property
    def id(self):
        return self.get_id()

    def get_id(self):
        return None

    @property
    def roles(self):
        return set(Role.query.all())

    @roles.setter
    def roles(self, _):
        pass

    def get_auth_token(self):
        return None


def create_dev_user():
    for user in SEED_USERS:
        user = user.copy()
        if users.get_user(user["email"]):
            continue

        roles = [users.find_or_create_role(r) for r in user.pop("_roles")]
        user["password"] = hash_password(user["password"])
        users.create_user(**user, roles=roles)

    db.session.commit()



from abc import ABC, abstractmethod


class IncompatibleVersionError(Exception):
    """Occurs when a component is incompatible with its representation"""

    def __init__(self, message, backend_version: int, version: int):
        super().__init__(message)

        self.backend_version = backend_version
        self.version = version


class Versioned(ABC):
    """Mixin to represent something that must be compatible with a certain version"""

    @property
    @abstractmethod
    def backend_version(self) -> int:
        pass

    def is_compatible(self, version: int = None):
        try:
            self.ensure_compatible(version=version)
            return True
        except IncompatibleVersionError:
            return False

    def ensure_compatible(self, version: int = None):
        version = self.__class__.__version__ if version is None else version
        backend_version = self.backend_version

        if backend_version != version:
            raise IncompatibleVersionError(
                f"Version {version} required, currently at {self.backend_version}",
                backend_version,
                version,
            )



import os
import threading
import subprocess

from meltano.core.project import Project
from meltano.core.db import project_engine
from meltano.core.utils.pidfile import PIDFile


class APIWorker(threading.Thread):
    def __init__(self, project: Project, bind_addr, reload=False):
        super().__init__()
        self.project = project
        self.bind_addr = bind_addr
        self.reload = reload
        self.pid_file = PIDFile(self.project.run_dir("gunicorn.pid"))

    def run(self):
        # fmt: off
        cmd = ["gunicorn",
               "--bind", self.bind_addr,
               "--config", "python:meltano.api.wsgi",
               "--pid", str(self.pid_file)]
        # fmt: on

        if self.reload:
            cmd += ["--reload"]

        cmd += ["meltano.api.app:create_app()"]

        engine, _ = project_engine(self.project)
        subprocess.run(cmd, env={**os.environ, "MELTANO_DATABASE_URI": str(engine.url)})

    def pid_path(self):
        return self.project.run_dir(f"gunicorn.pid")

    def stop(self):
        self.pid_file.process.terminate()



from typing import Callable
from functools import wraps


class visit_with:
    def __init__(self, visit: Callable):
        self.visit = visit

    def __call__(self, cls):
        # keep a reference to the proper visitor
        __visit__ = self.visit

        class Visitor(cls):
            def visit(self, node, *args, **kwargs):
                return __visit__(node, self, *args, **kwargs)

        return Visitor



import os
import copy
import logging
from enum import Enum
from pathlib import Path

from .elt_context import ELTContext
from .plugin_discovery_service import PluginDiscoveryService


DEFAULT_ANALYZE_SCHEMA = "analytics"
MELTANO_SCHEMA_SUFFIX = os.getenv("MELTANO_SCHEMA_SUFFIX", "")
MELTANO_LOAD_SCHEMA_SUFFIX = os.getenv(
    "MELTANO_LOAD_SCHEMA_SUFFIX", MELTANO_SCHEMA_SUFFIX
)
MELTANO_ANALYZE_SCHEMA_SUFFIX = os.getenv(
    "MELTANO_ANALYZE_SCHEMA_SUFFIX", MELTANO_SCHEMA_SUFFIX
)


class DialectNotSupportedError(Exception):
    pass


class ConnectionService:
    def __init__(self, elt_context: ELTContext):
        self.context = elt_context

    @property
    def dialect(self):
        return self.context.loader.namespace

    def load_params(self):
        schema = (
            os.getenv("MELTANO_LOAD_SCHEMA", self.context.extractor.namespace)
            + MELTANO_LOAD_SCHEMA_SUFFIX
        )

        dialect_params = {
            "postgres": {"schema": schema},
            "snowflake": {"schema": schema.upper()},
        }

        return dialect_params.get(self.dialect, {})

    def analyze_params(self):
        schema = (
            os.getenv("MELTANO_ANALYZE_SCHEMA", DEFAULT_ANALYZE_SCHEMA)
            + MELTANO_ANALYZE_SCHEMA_SUFFIX
        )

        dialect_params = {
            "postgres": {"schema": schema},
            "snowflake": {"schema": schema.upper()},
        }

        return dialect_params.get(self.dialect, {})

    def load_uri(self):
        params = self.load_params()
        return self.dialect_engine_uri(params)

    def analyze_uri(self):
        params = self.analyze_params()
        return self.dialect_engine_uri(params)

    def dialect_engine_uri(self, params={}):
        # this part knows probably too much about the setting definition
        # for a plugin, but we have to do this if we want to infer connections
        # from the loader
        db_suffix = lambda s: str(Path(s).with_suffix(".db"))

        dialect_templates = {
            "postgres": lambda params: "postgresql://{user}:{password}@{host}:{port}/{dbname}".format(
                **params
            ),
            "snowflake": lambda params: "snowflake://{username}:{password}@{account}/{database}/{schema}?warehouse={warehouse}&role={role}".format(
                **params
            ),
            "sqlite": lambda params: "sqlite:///{database}".format(
                database=db_suffix(params.pop("database")), **params
            ),
        }

        try:
            url = dialect_templates[self.dialect](
                {**self.context.loader.config, **params}
            )

            return url
        except KeyError:
            raise DialectNotSupportedError(self.dialect)



import logging
import requests
import uuid
import yaml
import os
import json

from typing import Dict

from meltano.core.utils import truthy

REQUEST_TIMEOUT = 2.0
MELTANO_TRACKING_ID = "UA-132758957-3"
MEASUREMENT_PROTOCOL_URI = "https://www.google-analytics.com/collect"
DEBUG_MEASUREMENT_PROTOCOL_URI = "https://www.google-analytics.com/debug/collect"


class GoogleAnalyticsTracker:
    def __init__(self, project, tracking_id: str = None, request_timeout: float = None):
        self.project = project
        self.tracking_id = tracking_id or MELTANO_TRACKING_ID
        self.request_timeout = request_timeout or REQUEST_TIMEOUT

        self.send_anonymous_usage_stats = (
            not truthy(os.getenv("MELTANO_DISABLE_TRACKING"))
            and self.project.meltano.get("send_anonymous_usage_stats", False) == True
        )
        self.project_id = self.load_project_id()
        self.client_id = self.load_client_id()

    def update_permission_to_track(self, send_anonymous_usage_stats: bool) -> None:
        """Update the send_anonymous_usage_stats in meltano.yml."""
        with self.project.meltano_update() as meltano_yml:
            meltano_yml["send_anonymous_usage_stats"] = send_anonymous_usage_stats

    def load_project_id(self) -> uuid.UUID:
        """
        Fetch the project_id from the project config file.

        If it is not found (e.g. first time run), generate a valid uuid4 and
        store it in the project config file.
        """
        try:
            project_id_str = self.project.meltano.get("project_id", None) or ""
            project_id = uuid.UUID(project_id_str, version=4)
        except ValueError:
            project_id = uuid.uuid4()

            if self.send_anonymous_usage_stats:
                # If we are set to track Anonymous Usage stats, also store
                #  the generated project_id back to the project config file
                #  so that it persists between meltano runs.
                with self.project.meltano_update() as meltano_yml:
                    meltano_yml["project_id"] = str(project_id)

        return project_id

    def load_client_id(self) -> uuid.UUID:
        """
        Fetch the client_id from the non-versioned analytics.json.

        If it is not found (e.g. first time run), generate a valid uuid4 and
        store it in analytics.json.
        """
        file_path = self.project.meltano_dir().joinpath("analytics.json")
        try:
            with open(file_path) as file:
                file_data = json.load(file)
            client_id_str = file_data["client_id"]
            client_id = uuid.UUID(client_id_str, version=4)
        except FileNotFoundError:
            client_id = uuid.uuid4()

            if self.send_anonymous_usage_stats:
                # If we are set to track Anonymous Usage stats, also store
                #  the generated client_id in a non-versioned analytics.json file
                #  so that it persists between meltano runs.
                with open(file_path, "w") as file:
                    data = {"client_id": str(client_id)}
                    json.dump(data, file)

        return client_id

    def event(self, category: str, action: str) -> Dict:
        """Constract a GA event with all the required parameters."""
        event = {
            "v": "1",
            "tid": self.tracking_id,
            "cid": self.client_id,
            "ds": "meltano cli",
            "t": "event",
            "ec": category,
            "ea": action,
            "el": self.project_id,
            "cd1": self.project_id,  # maps to the custom dimension 1 of the UI
        }
        return event

    def track_data(self, data: Dict, debug: bool = False) -> None:
        """Send usage statistics back to Google Analytics."""
        if self.send_anonymous_usage_stats == False:
            # Only send anonymous usage stats if you have explicit permission
            return

        if debug:
            tracking_uri = DEBUG_MEASUREMENT_PROTOCOL_URI
        else:
            tracking_uri = MEASUREMENT_PROTOCOL_URI

        try:
            r = requests.post(tracking_uri, data=data, timeout=self.request_timeout)

            if debug:
                logging.debug("GoogleAnalyticsTracker.track_data:")
                logging.debug(data)
                logging.debug("Response:")
                logging.debug(f"status_code: {r.status_code}")
                logging.debug(r.text)
        except requests.exceptions.Timeout:
            logging.debug("GoogleAnalyticsTracker.track_data: Request Timed Out")
        except requests.exceptions.ConnectionError as e:
            logging.debug("GoogleAnalyticsTracker.track_data: ConnectionError")
            logging.debug(e)
        except requests.exceptions.RequestException as e:
            logging.debug("GoogleAnalyticsTracker.track_data: RequestException")
            logging.debug(e)

    def track_event(self, category: str, action: str, debug: bool = False) -> None:
        event = self.event(category, action)
        self.track_data(event, debug)

    def track_meltano_init(self, project_name: str, debug: bool = False) -> None:
        event = self.track_event(
            category="meltano init", action=f"meltano init {project_name}", debug=debug
        )

    def track_meltano_add(
        self, plugin_type: str, plugin_name: str, debug: bool = False
    ) -> None:
        event = self.track_event(
            category=f"meltano add {plugin_type}",
            action=f"meltano add {plugin_type} {plugin_name}",
            debug=debug,
        )

    def track_meltano_discover(self, plugin_type: str, debug: bool = False) -> None:
        event = self.track_event(
            category="meltano discover",
            action=f"meltano discover {plugin_type}",
            debug=debug,
        )

    def track_meltano_elt(
        self, extractor: str, loader: str, transform: str, debug: bool = False
    ) -> None:
        event = self.track_event(
            category="meltano elt",
            action=f"meltano elt {extractor} {loader} --transform {transform}",
            debug=debug,
        )

    def track_meltano_install(self, debug: bool = False) -> None:
        event = self.track_event(
            category="meltano install", action="meltano install", debug=debug
        )

    def track_meltano_invoke(
        self, plugin_name: str, plugin_args: str, debug: bool = False
    ) -> None:
        event = self.track_event(
            category="meltano invoke",
            action=f"meltano invoke {plugin_name} {plugin_args}",
            debug=debug,
        )

    def track_meltano_schedule(self, schedule, debug=False) -> None:
        self.track_event(
            category="meltano schedule",
            action=f"meltano schedule {schedule.name} {schedule.extractor} {schedule.loader} {schedule.interval} --transform={schedule.transform}",
            debug=debug,
        )

    def track_meltano_permissions_grant(
        self, db: str, dry: bool, debug: bool = False
    ) -> None:
        action = f"meltano permissions grant --db {db}"
        if dry:
            action = action + " --dry"

        event = self.track_event(
            category="meltano permissions grant", action=action, debug=debug
        )

    def track_meltano_select(
        self,
        extractor: str,
        entities_filter: str,
        attributes_filter: str,
        flags: [],
        debug: bool = False,
    ) -> None:
        action = f"meltano select {extractor} {entities_filter} {attributes_filter}"

        if flags["list"]:
            action = action + " --list"
        if flags["all"]:
            action = action + " --all"
        if flags["exclude"]:
            action = action + " --exclude"

        event = self.track_event(category="meltano select", action=action, debug=debug)

    def track_meltano_ui(self, debug: bool = False) -> None:
        action = f"meltano ui"
        event = self.track_event(category="meltano ui", action=action, debug=debug)



import cerberus
import click
import logging
import yaml
import re

from typing import Dict, List, Tuple

from meltano.core.permissions.utils.error import SpecLoadingError
from meltano.core.permissions.utils.snowflake_connector import SnowflakeConnector
from meltano.core.permissions.spec_schemas.snowflake import *
from meltano.core.permissions.utils.snowflake_grants import SnowflakeGrantsGenerator
from meltano.core.permissions.utils.snowflake_revoke import SnowflakeRevokesGenerator


VALIDATION_ERR_MSG = 'Spec error: {} "{}", field "{}": {}'


class SnowflakeSpecLoader:
    def __init__(self, spec_path: str) -> None:
        # Load the specification file and check for (syntactical) errors
        click.secho("Loading spec file", fg="green")
        self.spec = self.load_spec(spec_path)

        # Generate the entities (e.g databases, schemas, users, etc) referenced
        #  by the spec file and make sure that no syntatical or refernce errors
        #  exist (all referenced entities are also defined by the spec)
        click.secho("Checking spec file for errors", fg="green")
        self.entities = self.inspect_spec()

        # Connect to Snowflake to make sure that all entities defined in the
        #  spec file are also defined in Snowflake (no missing databases, etc)
        click.secho(
            "Checking that all entities in the spec file are defined in Snowflake",
            fg="green",
        )
        self.check_entities_on_snowflake_server()

        # Get the privileges granted to users and roles in the Snowflake account
        # Used in order to figure out which permissions in the spec file are
        #  new ones and which already exist (and there is no need to re-grant them)
        click.secho("Fetching granted privileges from Snowflake", fg="green")
        self.grants_to_role = {}
        self.roles_granted_to_user = {}
        self.get_privileges_from_snowflake_server()

    def load_spec(self, spec_path: str) -> Dict:
        """
        Load a permissions specification from a file.

        If the file is not found or at least an error is found during validation,
        raise a SpecLoadingError with the appropriate error messages.

        Otherwise, return the valid specification as a Dictionary to be used
        in other operations.

        Raises a SpecLoadingError with all the errors found in the spec if at
        least one error is found.

        Returns the spec as a dictionary if everything is OK
        """
        try:
            with open(spec_path, "r") as stream:
                spec = yaml.load(stream)
        except FileNotFoundError:
            raise SpecLoadingError(f"Spec File {spec_path} not found")

        error_messages = self.ensure_valid_schema(spec)
        if error_messages:
            raise SpecLoadingError("\n".join(error_messages))

        return spec

    def ensure_valid_schema(self, spec: Dict) -> List[str]:
        """
        Ensure that the provided spec has no schema errors.

        Returns a list with all the errors found.
        """
        error_messages = []

        validator = cerberus.Validator(yaml.load(SNOWFLAKE_SPEC_SCHEMA))
        validator.validate(spec)
        for entity_type, err_msg in validator.errors.items():
            if isinstance(err_msg[0], str):
                error_messages.append(f"Spec error: {entity_type}: {err_msg[0]}")
                continue

            for error in err_msg[0].values():
                error_messages.append(f"Spec error: {entity_type}: {error[0]}")

        if error_messages:
            return error_messages

        schema = {
            "databases": yaml.load(SNOWFLAKE_SPEC_DATABASE_SCHEMA),
            "roles": yaml.load(SNOWFLAKE_SPEC_ROLE_SCHEMA),
            "users": yaml.load(SNOWFLAKE_SPEC_USER_SCHEMA),
            "warehouses": yaml.load(SNOWFLAKE_SPEC_WAREHOUSE_SCHEMA),
        }

        validators = {
            "databases": cerberus.Validator(schema["databases"]),
            "roles": cerberus.Validator(schema["roles"]),
            "users": cerberus.Validator(schema["users"]),
            "warehouses": cerberus.Validator(schema["warehouses"]),
        }

        entities_by_type = [
            (entity_type, entities)
            for entity_type, entities in spec.items()
            if entities and entity_type != "version"
        ]

        for entity_type, entities in entities_by_type:
            for entity_dict in entities:
                for entity_name, config in entity_dict.items():
                    validators[entity_type].validate(config)
                    for field, err_msg in validators[entity_type].errors.items():
                        error_messages.append(
                            VALIDATION_ERR_MSG.format(
                                entity_type, entity_name, field, err_msg[0]
                            )
                        )

        return error_messages

    def inspect_spec(self) -> Dict:
        """
        Inspect a valid spec and make sure that no logic errors exist.

        e.g. a role granted to a user not defined in roles
             or a user given acces to a database not defined in databases

        If at least an error is found during inspection, raise a
        SpecLoadingError with the appropriate error messages.

        Otherwise, return the entities found as a Dictionary to be used
        in other operations.

        Raises a SpecLoadingError with all the errors found in the spec if at
        least an error is found.

        Returns a dictionary with all the entities defined in the spec
        """
        entities, error_messages = self.generate_entities()

        error_messages.extend(self.ensure_valid_entity_names(entities))

        error_messages.extend(self.ensure_valid_references(entities))

        if error_messages:
            raise SpecLoadingError("\n".join(error_messages))

        return entities

    def generate_entities(self) -> Tuple[Dict, List[str]]:
        """
        Generate and return a dictionary with all the entities defined or
        referenced in the permissions specification file.

        The xxx_refs entities are referenced by various permissions.
        For example:
        'roles' --> All the roles defined in the spec
        'role_refs' --> All the roles referenced in a member_of permission
        'table_refs' --> All the tables referenced in read/write privileges
                         or in owns entries

        Returns a tuple (entities, error_messages) with all the entities defined
        in the spec and any errors found (e.g. a user not assigned their user role)
        """
        error_messages = []

        entities = {
            "databases": set(),
            "database_refs": set(),
            "shared_databases": set(),
            "schema_refs": set(),
            "table_refs": set(),
            "roles": set(),
            "role_refs": set(),
            "users": set(),
            "warehouses": set(),
            "warehouse_refs": set(),
        }

        entities_by_type = [
            (entity_type, entry)
            for entity_type, entry in self.spec.items()
            if entry and entity_type != "version"
        ]

        for entity_type, entry in entities_by_type:
            for entity_dict in entry:
                for entity_name, config in entity_dict.items():
                    if entity_type == "databases":
                        entities["databases"].add(entity_name)

                        if "shared" in config:
                            if type(config["shared"]) == bool:
                                if config["shared"]:
                                    entities["shared_databases"].add(entity_name)
                            else:
                                logging.debug(
                                    "`shared` for database {} must be boolean, skipping Role Reference generation.".format(
                                        entity_name
                                    )
                                )

                    elif entity_type == "roles":
                        entities["roles"].add(entity_name)

                        try:
                            for member_role in config["member_of"]:
                                entities["role_refs"].add(member_role)
                        except KeyError:
                            logging.debug(
                                "`member_of` not found for role {}, skipping Role Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for warehouse in config["warehouses"]:
                                entities["warehouse_refs"].add(warehouse)
                        except KeyError:
                            logging.debug(
                                "`warehouses` not found for role {}, skipping Warehouse Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for schema in config["privileges"]["databases"]["read"]:
                                entities["database_refs"].add(schema)
                        except KeyError:
                            logging.debug(
                                "`privileges.databases.read` not found for role {}, skipping Database Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for schema in config["privileges"]["databases"]["write"]:
                                entities["database_refs"].add(schema)
                        except KeyError:
                            logging.debug(
                                "`privileges.databases.write` not found for role {}, skipping Database Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for schema in config["privileges"]["schemas"]["read"]:
                                entities["schema_refs"].add(schema)
                        except KeyError:
                            logging.debug(
                                "`privileges.schemas.read` not found for role {}, skipping Schema Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for schema in config["privileges"]["schemas"]["write"]:
                                entities["schema_refs"].add(schema)
                        except KeyError:
                            logging.debug(
                                "`privileges.schemas.write` not found for role {}, skipping Schema Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for table in config["privileges"]["tables"]["read"]:
                                entities["table_refs"].add(table)
                        except KeyError:
                            logging.debug(
                                "`privileges.tables.read` not found for role {}, skipping Table Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for table in config["privileges"]["tables"]["write"]:
                                entities["table_refs"].add(table)
                        except KeyError:
                            logging.debug(
                                "`privileges.tables.write` not found for role {}, skipping Table Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for schema in config["owns"]["databases"]:
                                entities["database_refs"].add(schema)
                        except KeyError:
                            logging.debug(
                                "`owns.databases` not found for role {}, skipping Database Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for schema in config["owns"]["schemas"]:
                                entities["schema_refs"].add(schema)
                        except KeyError:
                            logging.debug(
                                "`owns.schemas` not found for role {}, skipping Schema Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for table in config["owns"]["tables"]:
                                entities["table_refs"].add(table)
                        except KeyError:
                            logging.debug(
                                "`owns.tables` not found for role {}, skipping Table Reference generation.".format(
                                    entity_name
                                )
                            )

                    elif entity_type == "users":
                        entities["users"].add(entity_name)

                        try:
                            for member_role in config["member_of"]:
                                entities["role_refs"].add(member_role)
                        except KeyError:
                            logging.debug(
                                "`member_of` not found for user {}, skipping Role Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for schema in config["owns"]["databases"]:
                                entities["database_refs"].add(schema)
                        except KeyError:
                            logging.debug(
                                "`owns.databases` not found for user {}, skipping Database Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for schema in config["owns"]["schemas"]:
                                entities["schema_refs"].add(schema)
                        except KeyError:
                            logging.debug(
                                "`owns.schemas` not found for user {}, skipping Schema Reference generation.".format(
                                    entity_name
                                )
                            )

                        try:
                            for table in config["owns"]["tables"]:
                                entities["table_refs"].add(table)
                        except KeyError:
                            logging.debug(
                                "`owns.tables` not found for user {}, skipping Table Reference generation.".format(
                                    entity_name
                                )
                            )

                    elif entity_type == "warehouses":
                        entities["warehouses"].add(entity_name)

        # Add implicit references to DBs and Schemas.
        #  e.g. RAW.MYSCHEMA.TABLE references also DB RAW and Schema MYSCHEMA
        for schema in entities["schema_refs"]:
            name_parts = schema.split(".")
            # Add the Database in the database refs
            if name_parts[0] != "*":
                entities["database_refs"].add(name_parts[0])

        for table in entities["table_refs"]:
            name_parts = table.split(".")
            # Add the Database in the database refs
            if name_parts[0] != "*":
                entities["database_refs"].add(name_parts[0])

            # Add the Schema in the schema refs
            if name_parts[1] != "*":
                entities["schema_refs"].add(f"{name_parts[0]}.{name_parts[1]}")

        return (entities, error_messages)

    def ensure_valid_entity_names(self, entities: Dict) -> List[str]:
        """
        Check that all entity names are valid.

        Returns a list with all the errors found.
        """
        error_messages = []

        for db in entities["databases"].union(entities["database_refs"]):
            name_parts = db.split(".")
            if not len(name_parts) == 1:
                error_messages.append(
                    f"Name error: Not a valid database name: {db}"
                    " (Proper definition: DB)"
                )

        for schema in entities["schema_refs"]:
            name_parts = schema.split(".")
            if (not len(name_parts) == 2) or (name_parts[0] == "*"):
                error_messages.append(
                    f"Name error: Not a valid schema name: {schema}"
                    " (Proper definition: DB.[SCHEMA | *])"
                )

        for table in entities["table_refs"]:
            name_parts = table.split(".")
            if (not len(name_parts) == 3) or (name_parts[0] == "*"):
                error_messages.append(
                    f"Name error: Not a valid table name: {table}"
                    " (Proper definition: DB.[SCHEMA | *].[TABLE | *])"
                )
            elif name_parts[1] == "*" and name_parts[2] != "*":
                error_messages.append(
                    f"Name error: Not a valid table name: {table}"
                    " (Can't have a Table name after selecting all schemas"
                    " with *: DB.SCHEMA.[TABLE | *])"
                )

        return error_messages

    def ensure_valid_references(self, entities: Dict) -> List[str]:
        """
        Make sure that all references are well defined.

        Returns a list with all the errors found.
        """
        error_messages = []

        # Check that all the referenced entities are also defined
        for database in entities["database_refs"]:
            if database not in entities["databases"]:
                error_messages.append(
                    f"Reference error: Database {database} is referenced "
                    "in the spec but not defined"
                )

        for role in entities["role_refs"]:
            if role not in entities["roles"]:
                error_messages.append(
                    f"Reference error: Role {role} is referenced in the "
                    "spec but not defined"
                )

        for warehouse in entities["warehouse_refs"]:
            if warehouse not in entities["warehouses"]:
                error_messages.append(
                    f"Reference error: Warehouse {warehouse} is referenced "
                    "in the spec but not defined"
                )

        return error_messages

    def check_entities_on_snowflake_server(self) -> None:
        """
        Make sure that all [warehouses, dbs, schemas, tables, users, roles]
        referenced in the spec are defined in Snowflake.

        Raises a SpecLoadingError with all the errors found while checking
        Snowflake for missinf entities.
        """
        error_messages = []

        conn = SnowflakeConnector()

        warehouses = conn.show_warehouses()
        for warehouse in self.entities["warehouses"]:
            if warehouse.upper() not in warehouses:
                error_messages.append(
                    f"Missing Entity Error: Warehouse {warehouse} was not found on"
                    " Snowflake Server. Please create it before continuing."
                )

        databases = conn.show_databases()
        for db in self.entities["databases"]:
            if db.upper() not in databases:
                error_messages.append(
                    f"Missing Entity Error: Database {db} was not found on"
                    " Snowflake Server. Please create it before continuing."
                )

        schemas = conn.show_schemas()
        for schema in self.entities["schema_refs"]:
            if "*" not in schema and schema.upper() not in schemas:
                error_messages.append(
                    f"Missing Entity Error: Schema {schema} was not found on"
                    " Snowflake Server. Please create it before continuing."
                )

        tables = conn.show_tables()
        views = conn.show_views()
        for table in self.entities["table_refs"]:
            if (
                "*" not in table
                and table.upper() not in tables
                and table.upper() not in views
            ):
                error_messages.append(
                    f"Missing Entity Error: Table/View {table} was not found on"
                    " Snowflake Server. Please create it before continuing."
                )

        roles = conn.show_roles()
        for role in self.entities["roles"]:
            if role.upper() not in roles:
                error_messages.append(
                    f"Missing Entity Error: Role {role} was not found on"
                    " Snowflake Server. Please create it before continuing."
                )

        users = conn.show_users()
        for user in self.entities["users"]:
            if user.upper() not in users:
                error_messages.append(
                    f"Missing Entity Error: User {user} was not found on"
                    " Snowflake Server. Please create it before continuing."
                )

        if error_messages:
            raise SpecLoadingError("\n".join(error_messages))

    def get_privileges_from_snowflake_server(self) -> None:
        """
        Get the privileges granted to users and roles in the Snowflake account
        """
        conn = SnowflakeConnector()

        for role in self.entities["roles"]:
            self.grants_to_role[role] = conn.show_grants_to_role(role)

        for user in self.entities["users"]:
            self.roles_granted_to_user[user] = conn.show_roles_granted_to_user(user)

    def generate_revoke_queries(self) -> List[Dict]:
        """
        Starting point to generate all the revoke queries.

        For each entity type (e.g. user or role) that is affected by the spec,
        the proper sql revoke queries are generated.

        Returns all the SQL commands as a list.
        """
        sql_commands = []

        generator = SnowflakeRevokesGenerator(
            self.grants_to_role, self.roles_granted_to_user
        )

        click.secho("Generating revoke statements for Snowflake", fg="green")

        for entity_type, entry in self.spec.items():
            if entity_type in ["databases", "warehouses", "version"]:
                continue

            for entity_dict in entry:
                entity_configs = [
                    (entity_name, config)
                    for entity_name, config in entity_dict.items()
                    if config
                ]

                for entity_name, config in entity_configs:
                    if entity_type == "roles":
                        click.secho(f"     Processing role {entity_name}", fg="green")
                        sql_commands.extend(
                            generator.generate_revoke_roles("ROLE", entity_name, config)
                        )

                        sql_commands.extend(
                            generator.generate_revoke_privileges_to_role(
                                entity_name, config, self.entities["shared_databases"]
                            )
                        )
                    elif entity_type == "users":
                        click.secho(f"     Processing user {entity_name}", fg="green")
                        sql_commands.extend(
                            generator.generate_revoke_roles("USER", entity_name, config)
                        )

        return self.remove_duplicate_queries(sql_commands)

    def generate_permission_queries(self) -> List[Dict]:
        """
        Starting point to generate all the permission queries.

        For each entity type (e.g. user or role) that is affected by the spec,
        the proper sql permission queries are generated.

        Returns all the SQL commands as a list.
        """
        sql_commands = []

        generator = SnowflakeGrantsGenerator(
            self.grants_to_role, self.roles_granted_to_user
        )

        click.secho("Generating permission Queries:", fg="green")

        # For each permission in the spec, check if we have to generate an
        #  SQL command granting that permission
        for entity_type, entry in self.spec.items():
            if entity_type in ["databases", "warehouses", "version"]:
                continue

            for entity_dict in entry:
                entity_configs = [
                    (entity_name, config)
                    for entity_name, config in entity_dict.items()
                    if config
                ]

                for entity_name, config in entity_configs:
                    if entity_type == "roles":
                        click.secho(f"     Processing role {entity_name}", fg="green")
                        sql_commands.extend(
                            generator.generate_grant_roles("ROLE", entity_name, config)
                        )

                        sql_commands.extend(
                            generator.generate_grant_ownership(entity_name, config)
                        )

                        sql_commands.extend(
                            generator.generate_grant_privileges_to_role(
                                entity_name, config, self.entities["shared_databases"]
                            )
                        )
                    elif entity_type == "users":
                        click.secho(f"     Processing user {entity_name}", fg="green")
                        sql_commands.extend(
                            generator.generate_alter_user(entity_name, config)
                        )

                        sql_commands.extend(
                            generator.generate_grant_roles("USER", entity_name, config)
                        )

        return self.remove_duplicate_queries(sql_commands)

    def remove_duplicate_queries(self, sql_commands: Dict) -> List[Dict]:
        grants = []
        revokes = []

        for i, command in reversed(list(enumerate(sql_commands))):
            # Find all "GRANT OWNERSHIP commands"
            if command["sql"].startswith("GRANT OWNERSHIP ON"):
                grant = (command["sql"].split("TO ROLE", 1)[0]).upper()

                if grant in grants:
                    # If there is already a GRANT OWNERSHIP for the same
                    #  DB/SCHEMA/TABLE --> remove the one before it
                    #  (only keep the last one)
                    del sql_commands[i]
                else:
                    grants.append(grant)

            if command["sql"].startswith("REVOKE ALL"):
                revoke = command["sql"].upper()
                if revoke in revokes:
                    # If there is already a REVOKE ALL for the same
                    #  DB/SCHEMA/TABLE --> remove the one before it
                    #  (only keep the last one)
                    del sql_commands[i]
                else:
                    revokes.append(revoke)

        return sql_commands



import click
from . import cli
from .params import project

from meltano.core.compiler.project_compiler import ProjectCompiler


@cli.group(hidden=True, invoke_without_command=True)
@project()
@click.pass_context
def model(ctx, project):
    compiler = ctx.obj["compiler"] = ProjectCompiler(project)
    compiler.parse()

    if ctx.invoked_subcommand is None:
        click.secho("Local", fg="green")
        for model in compiler.topics:
            click.echo(f'{model["namespace"]}/{model["name"]}')

        print()

        click.secho("Packaged", fg="green")
        for model in compiler.package_topics:
            click.echo(f'{model["namespace"]}/{model["name"]}')


@model.command()
@click.pass_context
def compile(ctx):
    compiler = ctx.obj["compiler"]

    compiler.compile()
    click.secho(f"Compiled {len(compiler.topics + compiler.package_topics)} model(s)")



from pathlib import Path
import sqlalchemy

# the `scripts/alembic_freeze.py` refers to these paths
# we need to make sure they match
MIGRATION_DIR = Path(__path__[0])
LOCK_PATH = MIGRATION_DIR.joinpath("db.lock")


from meltano.api.models.security import Role, RolePermissions
from meltano.core.db import project_engine


def seed(project):
    _, Session = project_engine(project)

    try:
        session = Session()

        if not session.query(Role).filter_by(name="admin").first():
            session.add(
                Role(
                    name="admin",
                    description="Meltano Admin",
                    permissions=[
                        RolePermissions(type="view:design", context="*"),
                        RolePermissions(type="view:reports", context="*"),
                        RolePermissions(type="modify:acl", context="*"),
                    ],
                )
            )

        if not session.query(Role).filter_by(name="regular").first():
            session.merge(Role(name="regular", description="Meltano User"))

        # add the universal permissions to Admin
        admin = session.query(Role).filter_by(name="admin").one()
        try:
            session.query(RolePermissions).filter_by(
                role=admin, type="*", context="*"
            ).one()
        except sqlalchemy.orm.exc.NoResultFound:
            admin.permissions.append(RolePermissions(type="*", context="*"))

        session.commit()
    finally:
        session.close()



import logging

from typing import Dict, List, Tuple, Set

from meltano.core.permissions.utils.snowflake_connector import SnowflakeConnector


GRANT_ROLE_TEMPLATE = "GRANT ROLE {role_name} TO {type} {entity_name}"

GRANT_PRIVILEGES_TEMPLATE = (
    "GRANT {privileges} ON {resource_type} {resource_name} TO ROLE {role}"
)

GRANT_ALL_PRIVILEGES_TEMPLATE = "GRANT {privileges} ON ALL {resource_type}S IN SCHEMA {resource_name} TO ROLE {role}"

GRANT_FUTURE_PRIVILEGES_TEMPLATE = "GRANT {privileges} ON FUTURE {resource_type}S IN SCHEMA {resource_name} TO ROLE {role}"

ALTER_USER_TEMPLATE = "ALTER USER {user_name} SET {privileges}"

GRANT_OWNERSHIP_TEMPLATE = (
    "GRANT OWNERSHIP"
    " ON {resource_type} {resource_name}"
    " TO ROLE {role_name} COPY CURRENT GRANTS"
)


class SnowflakeGrantsGenerator:
    def __init__(self, grants_to_role: Dict, roles_granted_to_user: Dict) -> None:
        self.grants_to_role = grants_to_role
        self.roles_granted_to_user = roles_granted_to_user

    def check_grant_to_role(
        self, role: str, privilege: str, entity_type: str, entity_name: str
    ) -> bool:
        """
        Check if <role> has been granted the privilege <privilege> on entity type
        <entity_type> with name <entity_name>.

        For example:
        check_grant_to_role('reporter', 'USAGE', 'DATABASE', 'ANALYTICS') -> True
        means that role reported has been granted the privilege to use the
        Database ANALYTICS on the Snowflake server.
        """
        if SnowflakeConnector.snowflaky(entity_name).upper() in self.grants_to_role.get(
            role, {}
        ).get(privilege, {}).get(entity_type, []):
            return True
        else:
            return False

    def generate_grant_roles(
        self, entity_type: str, entity: str, config: str
    ) -> List[Dict]:
        """
        Generate the GRANT statements for both roles and users.

        entity_type: "USER" or "ROLE"
        entity: the name of the entity (e.g. "yannis" or "REPORTER")
        config: the subtree for the entity as specified in the spec

        Returns the SQL commands generated as a list
        """
        sql_commands = []

        try:
            for member_role in config["member_of"]:
                granted_role = SnowflakeConnector.snowflaky(member_role).upper()
                already_granted = False
                if (
                    entity_type == "USER"
                    and granted_role in self.roles_granted_to_user[entity]
                ) or (
                    entity_type == "ROLE"
                    and self.check_grant_to_role(entity, "USAGE", "ROLE", member_role)
                ):
                    already_granted = True

                sql_commands.append(
                    {
                        "already_granted": already_granted,
                        "sql": GRANT_ROLE_TEMPLATE.format(
                            role_name=SnowflakeConnector.snowflaky(member_role),
                            type=entity_type,
                            entity_name=SnowflakeConnector.snowflaky(entity),
                        ),
                    }
                )
        except KeyError:
            logging.debug(
                "`member_of` not found for {}, skipping generation of GRANT ROLE statements.".format(
                    entity
                )
            )

        return sql_commands

    def generate_grant_privileges_to_role(
        self, role: str, config: str, shared_dbs: Set
    ) -> List[Dict]:
        """
        Generate all the privilege granting statements for a role.

        Most of the SQL command that will be generated are privileges granted to
        roles and this function orchestrates the whole process

        role: the name of the role (e.g. "LOADER" or "REPORTER") the privileges
              are GRANTed to
        config: the subtree for the role as specified in the spec
        shared_dbs: a set of all the shared databases defined in the spec.
                    Used down the road by generate_database_grants() to also grant
                    "IMPORTED PRIVILEGES" when access is granted to a shared DB.

        Returns the SQL commands generated as a list
        """
        sql_commands = []

        # Track all the DBs and schemas that have been given access to (GRANT USAGE)
        # the given role. Used in order to recursively grant the required access
        # to DBs or schemas implicetly reference when permissions are GRANTED for a
        # child Schema or Table.
        # Example: A role is given read access to table MY_DB.MY_SCHEMA.MY_TABLE
        # 1. In order to access MY_TABLE, the role has to be able to access MY_DB.MY_SCHEMA
        # 2. The script checks if USAGE on MY_DB has been granted to the role and
        #    assigns it to the role if not (and adds the DB to usage_granted["databases"])
        # 3. The same for the schema MY_SCHEMA
        # 4. Finaly the requested permissions are GRANTED to role for MY_TABLE
        usage_granted = {"databases": set(), "schemas": set()}

        try:
            for warehouse in config["warehouses"]:
                new_commands = self.generate_warehouse_grants(
                    role=role, warehouse=warehouse
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`warehouses` not found for role {}, skipping generation of Warehouse GRANT statements.".format(
                    role
                )
            )

        try:
            for database in config["privileges"]["databases"]["read"]:
                new_commands, usage_granted = self.generate_database_grants(
                    role=role,
                    database=database,
                    grant_type="read",
                    usage_granted=usage_granted,
                    shared_dbs=shared_dbs,
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.databases.read` not found for role {}, skipping generation of DATABASE read level GRANT statements.".format(
                    role
                )
            )

        try:
            for database in config["privileges"]["databases"]["write"]:
                new_commands, usage_granted = self.generate_database_grants(
                    role=role,
                    database=database,
                    grant_type="write",
                    usage_granted=usage_granted,
                    shared_dbs=shared_dbs,
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.databases.write` not found for role {}, skipping generation of DATABASE write level GRANT statements.".format(
                    role
                )
            )

        try:
            for schema in config["privileges"]["schemas"]["read"]:
                new_commands, usage_granted = self.generate_schema_grants(
                    role=role,
                    schema=schema,
                    grant_type="read",
                    usage_granted=usage_granted,
                    shared_dbs=shared_dbs,
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.schemas.read` not found for role {}, skipping generation of SCHEMA read level GRANT statements.".format(
                    role
                )
            )

        try:
            for schema in config["privileges"]["schemas"]["write"]:
                new_commands, usage_granted = self.generate_schema_grants(
                    role=role,
                    schema=schema,
                    grant_type="write",
                    usage_granted=usage_granted,
                    shared_dbs=shared_dbs,
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.schemas.write` not found for role {}, skipping generation of SCHEMA write level GRANT statements.".format(
                    role
                )
            )

        try:
            for table in config["privileges"]["tables"]["read"]:
                new_commands, usage_granted = self.generate_table_and_view_grants(
                    role=role,
                    table=table,
                    grant_type="read",
                    usage_granted=usage_granted,
                    shared_dbs=shared_dbs,
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.tables.read` not found for role {}, skipping generation of TABLE read level GRANT statements.".format(
                    role
                )
            )

        try:
            for table in config["privileges"]["tables"]["write"]:
                new_commands, usage_granted = self.generate_table_and_view_grants(
                    role=role,
                    table=table,
                    grant_type="write",
                    usage_granted=usage_granted,
                    shared_dbs=shared_dbs,
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.tables.write` not found for role {}, skipping generation of TABLE write level GRANT statements.".format(
                    role
                )
            )

        return sql_commands

    def generate_warehouse_grants(self, role: str, warehouse: str) -> List[str]:
        """
        Generate the GRANT statements for Warehouse usage and operation.

        role: the name of the role the privileges are GRANTed to
        warehouse: the name of the warehouse (e.g. "transforming")

        Returns the SQL command generated
        """
        sql_commands = []

        if self.check_grant_to_role(role, "USAGE", "WAREHOUSE", warehouse):
            already_granted = True
        else:
            already_granted = False

        sql_commands.append(
            {
                "already_granted": already_granted,
                "sql": GRANT_PRIVILEGES_TEMPLATE.format(
                    privileges="USAGE",
                    resource_type="WAREHOUSE",
                    resource_name=SnowflakeConnector.snowflaky(warehouse),
                    role=SnowflakeConnector.snowflaky(role),
                ),
            }
        )

        if self.check_grant_to_role(role, "OPERATE", "WAREHOUSE", warehouse):
            already_granted = True
        else:
            already_granted = False

        sql_commands.append(
            {
                "already_granted": already_granted,
                "sql": GRANT_PRIVILEGES_TEMPLATE.format(
                    privileges="OPERATE",
                    resource_type="WAREHOUSE",
                    resource_name=SnowflakeConnector.snowflaky(warehouse),
                    role=SnowflakeConnector.snowflaky(role),
                ),
            }
        )

        return sql_commands

    def generate_database_grants(
        self,
        role: str,
        database: str,
        grant_type: str,
        usage_granted: Dict,
        shared_dbs: Set,
    ) -> Tuple[List[str], Dict]:
        """
        Generate the GRANT statements for Databases.

        role: the name of the role the privileges are GRANTed to
        database: the name of the database (e.g. "RAW")
        grant_type: What type of privileges are granted? One of {"read", "write"}
        usage_granted: Passed by generate_grant_privileges_to_role() to track all
            all the entities a role has been granted access (USAGE) to.
        shared_dbs: a set of all the shared databases defined in the spec.

        Returns the SQL commands generated and the updated usage_granted as a Tuple
        """
        sql_commands = []

        usage_granted["databases"].add(database.upper())
        already_granted = False

        if grant_type == "read":
            privileges = "USAGE"

            if self.check_grant_to_role(role, "USAGE", "DATABASE", database):
                already_granted = True
        elif grant_type == "write":
            privileges = "USAGE, MONITOR, CREATE SCHEMA"

            if (
                self.check_grant_to_role(role, "USAGE", "DATABASE", database)
                and self.check_grant_to_role(role, "MONITOR", "DATABASE", database)
                and self.check_grant_to_role(
                    role, "CREATE SCHEMA", "DATABASE", database
                )
            ):
                already_granted = True
        else:
            raise SpecLoadingError(
                f"Wrong grant_type {spec_path} provided to generate_database_grants()"
            )

        # If this is a shared database, we have to grant the "IMPORTED PRIVILEGES"
        #  privilege to the user and skip granting the specific permissions as
        #  "Granting individual privileges on imported databases is not allowed."
        if database in shared_dbs:
            sql_commands.append(
                {
                    "already_granted": already_granted,
                    "sql": GRANT_PRIVILEGES_TEMPLATE.format(
                        privileges="IMPORTED PRIVILEGES",
                        resource_type="DATABASE",
                        resource_name=SnowflakeConnector.snowflaky(database),
                        role=SnowflakeConnector.snowflaky(role),
                    ),
                }
            )

            return (sql_commands, usage_granted)

        # And then grant privileges to the database
        sql_commands.append(
            {
                "already_granted": already_granted,
                "sql": GRANT_PRIVILEGES_TEMPLATE.format(
                    privileges=privileges,
                    resource_type="DATABASE",
                    resource_name=SnowflakeConnector.snowflaky(database),
                    role=SnowflakeConnector.snowflaky(role),
                ),
            }
        )

        return (sql_commands, usage_granted)

    def generate_schema_grants(
        self,
        role: str,
        schema: str,
        grant_type: str,
        usage_granted: Dict,
        shared_dbs: Set,
    ) -> Tuple[List[str], Dict]:
        """
        Generate the GRANT statements for SCHEMAs.

        role: the name of the role the privileges are GRANTed to
        schema: the name of the Schema (e.g. "RAW.PUBLIC")
        grant_type: What type of privileges are granted? One of {"read", "write"}
        usage_granted: Passed by generate_grant_privileges_to_role() to track all
            all the entities a role has been granted access (USAGE) to.
        shared_dbs: a set of all the shared databases defined in the spec.

        Returns the SQL commands generated and the updated usage_granted as a Tuple
        """
        sql_commands = []

        # Split the schema identifier into parts {DB_NAME}.{SCHEMA_NAME}
        #  so that we can check and use each one
        name_parts = schema.split(".")

        # Do nothing if this is a schema inside a shared database:
        #  "Granting individual privileges on imported databases is not allowed."
        if name_parts[0] in shared_dbs:
            return (sql_commands, usage_granted)

        if grant_type == "read":
            privileges = "USAGE"
        elif grant_type == "write":
            privileges = (
                "USAGE, MONITOR, CREATE TABLE,"
                " CREATE VIEW, CREATE STAGE, CREATE FILE FORMAT,"
                " CREATE SEQUENCE, CREATE FUNCTION, CREATE PIPE"
            )
        else:
            raise SpecLoadingError(
                f"Wrong grant_type {spec_path} provided to generate_schema_grants()"
            )

        # Before assigning privileges to a schema, check if
        #  usage to the database has been granted and
        # if not grant usage first to the database
        if name_parts[0].upper() not in usage_granted["databases"]:
            new_commands, usage_granted = self.generate_database_grants(
                role=role,
                database=name_parts[0],
                grant_type="read",
                usage_granted=usage_granted,
                shared_dbs=shared_dbs,
            )
            sql_commands.extend(new_commands)

        # Generate the INFORMATION_SCHEMA identifier for that database
        #  in order to be able to filter it out
        info_schema = f"{name_parts[0].upper()}.INFORMATION_SCHEMA"

        schemas = []

        if name_parts[1] == "*":
            # If {DB_NAME}.* was provided as the schema identifier, we have to fetch
            #  each schema in database DB_NAME, so that we can grant privileges
            #  for each schema seperatelly.
            # We could GRANT {privileges} TO ALL SCHEMAS IN DATABASE
            #  but that would not allow us to know if a specific privilege has
            #  been already granted or not
            conn = SnowflakeConnector()
            db_schemas = conn.show_schemas(name_parts[0])
            for db_schema in db_schemas:
                if db_schema != info_schema:
                    schemas.append(db_schema)
        elif "*" in name_parts[1]:
            conn = SnowflakeConnector()
            db_schemas = conn.show_schemas(name_parts[0])
            for db_schema in db_schemas:
                schema_name = db_schema.split(".", 1)[1].lower()
                if schema_name.startswith(name_parts[1].split("*", 1)[0]):
                    schemas.append(db_schema)
        else:
            schemas = [schema]

        for db_schema in schemas:
            already_granted = False

            if (
                grant_type == "read"
                and self.check_grant_to_role(role, "USAGE", "SCHEMA", db_schema)
            ) or (
                grant_type == "write"
                and self.check_grant_to_role(role, "USAGE", "SCHEMA", db_schema)
                and self.check_grant_to_role(role, "MONITOR", "SCHEMA", db_schema)
                and self.check_grant_to_role(role, "CREATE TABLE", "SCHEMA", db_schema)
                and self.check_grant_to_role(role, "CREATE VIEW", "SCHEMA", db_schema)
                and self.check_grant_to_role(role, "CREATE STAGE", "SCHEMA", db_schema)
                and self.check_grant_to_role(
                    role, "CREATE FILE FORMAT", "SCHEMA", db_schema
                )
                and self.check_grant_to_role(
                    role, "CREATE SEQUENCE", "SCHEMA", db_schema
                )
                and self.check_grant_to_role(
                    role, "CREATE FUNCTION", "SCHEMA", db_schema
                )
                and self.check_grant_to_role(role, "CREATE PIPE", "SCHEMA", db_schema)
            ):
                already_granted = True

            sql_commands.append(
                {
                    "already_granted": already_granted,
                    "sql": GRANT_PRIVILEGES_TEMPLATE.format(
                        privileges=privileges,
                        resource_type="SCHEMA",
                        resource_name=SnowflakeConnector.snowflaky(db_schema),
                        role=SnowflakeConnector.snowflaky(role),
                    ),
                }
            )

        usage_granted["schemas"].add(schema.upper())

        return (sql_commands, usage_granted)

    def generate_table_and_view_grants(
        self,
        role: str,
        table: str,
        grant_type: str,
        usage_granted: Dict,
        shared_dbs: Set,
    ) -> Tuple[List[str], Dict]:
        """
        Generate the GRANT statements for TABLEs and VIEWs.

        role: the name of the role the privileges are GRANTed to
        table: the name of the TABLE/VIEW (e.g. "RAW.PUBLIC.MY_TABLE")
        grant_type: What type of privileges are granted? One of {"read", "write"}
        usage_granted: Passed by generate_grant_privileges_to_role() to track all
            all the entities a role has been granted access (USAGE) to.
        shared_dbs: a set of all the shared databases defined in the spec.

        Returns the SQL commands generated and the updated usage_granted as a Tuple
        """
        sql_commands = []

        # Split the table identifier into parts {DB_NAME}.{SCHEMA_NAME}.{TABLE_NAME}
        #  so that we can check and use each one
        name_parts = table.split(".")

        # Do nothing if this is a table inside a shared database:
        #  "Granting individual privileges on imported databases is not allowed."
        if name_parts[0] in shared_dbs:
            return (sql_commands, usage_granted)

        if grant_type == "read":
            privileges = "SELECT"
        elif grant_type == "write":
            privileges = "SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES"
        else:
            raise SpecLoadingError(
                f"Wrong grant_type {spec_path} provided to generate_table_and_view_grants()"
            )

        # Generate the INFORMATION_SCHEMA identifier for that database
        #  in order to be able to filter it out
        info_schema = f"{name_parts[0].upper()}.INFORMATION_SCHEMA"

        # Before assigning privileges to a Table, check if
        #  usage to the database has been granted and
        # if not grant usage first to the database
        # Schemas will be checked as we generate them
        if name_parts[0].upper() not in usage_granted["databases"]:
            new_commands, usage_granted = self.generate_database_grants(
                role=role,
                database=name_parts[0],
                grant_type="read",
                usage_granted=usage_granted,
                shared_dbs=shared_dbs,
            )
            sql_commands.extend(new_commands)

        # Gather the tables/views that privileges will be granted to
        tables = []
        views = []

        # List of all tables/views in schema
        table_list = []
        view_list = []

        schemas = []

        conn = SnowflakeConnector()

        if name_parts[1] == "*":
            # If {DB_NAME}.*.* was provided as the identifier, we have to fetch
            #  each schema in database DB_NAME, so that we can grant privileges
            #  for all tables in that schema
            # (You can not GRANT to all table with a wild card for the schema name)
            db_schemas = conn.show_schemas(name_parts[0])
            for schema in db_schemas:
                if schema != info_schema:
                    schemas.append(schema)
        elif "*" in name_parts[1]:
            conn = SnowflakeConnector()
            db_schemas = conn.show_schemas(name_parts[0])
            for db_schema in db_schemas:
                schema_name = db_schema.split(".", 1)[1].lower()
                if schema_name.startswith(name_parts[1].split("*", 1)[0]):
                    schemas.append(db_schema)
        else:
            schemas = [f"{name_parts[0]}.{name_parts[1]}"]

        for schema in schemas:
            # first check if the role has been granted usage on the schema
            # either directly or indirectly (by granting to DB.*)
            if (
                schema.upper() not in usage_granted["schemas"]
                and f"{name_parts[0].upper()}.*" not in usage_granted["schemas"]
            ):
                new_commands, usage_granted = self.generate_schema_grants(
                    role=role,
                    schema=schema,
                    grant_type="read",
                    usage_granted=usage_granted,
                    shared_dbs=shared_dbs,
                )
                sql_commands.extend(new_commands)

            # And then add the tables for that schema to the tables[] and views[]
            #  that will be granted the permissions
            table_list.extend(conn.show_tables(schema=schema))
            view_list.extend(conn.show_views(schema=schema))

        if name_parts[2] == "*":
            # If *.* then you can grant all, grant future, and exit
            for schema in schemas:
                # Grant on ALL tables
                sql_commands.append(
                    {
                        "already_granted": False,
                        "sql": GRANT_ALL_PRIVILEGES_TEMPLATE.format(
                            privileges=privileges,
                            resource_type="TABLE",
                            resource_name=SnowflakeConnector.snowflaky(schema),
                            role=SnowflakeConnector.snowflaky(role),
                        ),
                    }
                )

                # Grant on ALL views
                sql_commands.append(
                    {
                        "already_granted": False,
                        "sql": GRANT_ALL_PRIVILEGES_TEMPLATE.format(
                            privileges=privileges,
                            resource_type="VIEW",
                            resource_name=SnowflakeConnector.snowflaky(schema),
                            role=SnowflakeConnector.snowflaky(role),
                        ),
                    }
                )

                # Grant future on all tables
                sql_commands.append(
                    {
                        "already_granted": False,
                        "sql": GRANT_FUTURE_PRIVILEGES_TEMPLATE.format(
                            privileges=privileges,
                            resource_type="TABLE",
                            resource_name=SnowflakeConnector.snowflaky(schema),
                            role=SnowflakeConnector.snowflaky(role),
                        ),
                    }
                )
                # Grant future on all views
                sql_commands.append(
                    {
                        "already_granted": False,
                        "sql": GRANT_FUTURE_PRIVILEGES_TEMPLATE.format(
                            privileges=privileges,
                            resource_type="VIEW",
                            resource_name=SnowflakeConnector.snowflaky(schema),
                            role=SnowflakeConnector.snowflaky(role),
                        ),
                    }
                )

            return (sql_commands, usage_granted)

        else:
            # Only one table/view to be granted permissions to
            if table.upper() in table_list:
                tables = [table]
            if table.upper() in view_list:
                views = [table]

        # And then grant privileges to all tables in that schema
        for db_table in tables:
            already_granted = False
            if (
                grant_type == "read"
                and self.check_grant_to_role(role, "SELECT", "TABLE", db_table)
            ) or (
                grant_type == "write"
                and self.check_grant_to_role(role, "SELECT", "TABLE", db_table)
                and self.check_grant_to_role(role, "INSERT", "TABLE", db_table)
                and self.check_grant_to_role(role, "UPDATE", "TABLE", db_table)
                and self.check_grant_to_role(role, "DELETE", "TABLE", db_table)
                and self.check_grant_to_role(role, "TRUNCATE", "TABLE", db_table)
                and self.check_grant_to_role(role, "REFERENCES", "TABLE", db_table)
            ):
                already_granted = True

            # And then grant privileges to the db_table
            sql_commands.append(
                {
                    "already_granted": already_granted,
                    "sql": GRANT_PRIVILEGES_TEMPLATE.format(
                        privileges=privileges,
                        resource_type="TABLE",
                        resource_name=SnowflakeConnector.snowflaky(db_table),
                        role=SnowflakeConnector.snowflaky(role),
                    ),
                }
            )

        for db_view in views:
            already_granted = False
            if (
                grant_type == "read"
                and self.check_grant_to_role(role, "SELECT", "VIEW", db_view)
            ) or (
                grant_type == "write"
                and self.check_grant_to_role(role, "SELECT", "VIEW", db_view)
                and self.check_grant_to_role(role, "INSERT", "VIEW", db_view)
                and self.check_grant_to_role(role, "UPDATE", "VIEW", db_view)
                and self.check_grant_to_role(role, "DELETE", "VIEW", db_view)
                and self.check_grant_to_role(role, "TRUNCATE", "VIEW", db_view)
                and self.check_grant_to_role(role, "REFERENCES", "VIEW", db_view)
            ):
                already_granted = True

            # And then grant privileges to the db_view
            sql_commands.append(
                {
                    "already_granted": already_granted,
                    "sql": GRANT_PRIVILEGES_TEMPLATE.format(
                        privileges=privileges,
                        resource_type="VIEW",
                        resource_name=SnowflakeConnector.snowflaky(db_view),
                        role=SnowflakeConnector.snowflaky(role),
                    ),
                }
            )

        return (sql_commands, usage_granted)

    def generate_alter_user(self, user: str, config: str) -> List[Dict]:
        """
        Generate the ALTER statements for USERs.

        user: the name of the USER
        config: the subtree for the user as specified in the spec

        Returns the SQL commands generated as a List
        """
        sql_commands = []
        alter_privileges = []

        if "can_login" in config:
            if config["can_login"]:
                alter_privileges.append("DISABLED = FALSE")
            else:
                alter_privileges.append("DISABLED = TRUE")

        if alter_privileges:
            sql_commands.append(
                {
                    "already_granted": False,
                    "sql": ALTER_USER_TEMPLATE.format(
                        user_name=SnowflakeConnector.snowflaky(user),
                        privileges=", ".join(alter_privileges),
                    ),
                }
            )

        return sql_commands

    def generate_grant_ownership(self, role: str, config: str) -> List[Dict]:
        """
        Generate the GRANT OWNERSHIP statements for DATABASEs, SCHEMAs and TABLEs.

        role: the name of the role (e.g. "LOADER") OWNERSHIP will be GRANTed to
        config: the subtree for the role as specified in the spec

        Returns the SQL commands generated as a List
        """
        sql_commands = []

        try:
            for database in config["owns"]["databases"]:
                if self.check_grant_to_role(role, "OWNERSHIP", "DATABASE", database):
                    already_granted = True
                else:
                    already_granted = False

                sql_commands.append(
                    {
                        "already_granted": already_granted,
                        "sql": GRANT_OWNERSHIP_TEMPLATE.format(
                            resource_type="DATABASE",
                            resource_name=SnowflakeConnector.snowflaky(database),
                            role_name=SnowflakeConnector.snowflaky(role),
                        ),
                    }
                )
        except KeyError:
            logging.debug(
                "`owns.databases` not found for role {}, skipping generation of DATABASE OWNERSHIP statements.".format(
                    role
                )
            )

        try:
            for schema in config["owns"]["schemas"]:
                name_parts = schema.split(".")
                info_schema = f"{name_parts[0].upper()}.INFORMATION_SCHEMA"

                schemas = []

                if name_parts[1] == "*":
                    conn = SnowflakeConnector()
                    db_schemas = conn.show_schemas(name_parts[0])

                    for db_schema in db_schemas:
                        if db_schema != info_schema:
                            schemas.append(db_schema)
                else:
                    schemas = [schema]

                for db_schema in schemas:
                    if self.check_grant_to_role(role, "OWNERSHIP", "SCHEMA", db_schema):
                        already_granted = True
                    else:
                        already_granted = False

                    sql_commands.append(
                        {
                            "already_granted": already_granted,
                            "sql": GRANT_OWNERSHIP_TEMPLATE.format(
                                resource_type="SCHEMA",
                                resource_name=SnowflakeConnector.snowflaky(db_schema),
                                role_name=SnowflakeConnector.snowflaky(role),
                            ),
                        }
                    )
        except KeyError:
            logging.debug(
                "`owns.schemas` not found for role {}, skipping generation of SCHEMA OWNERSHIP statements.".format(
                    role
                )
            )

        try:
            # Gather the tables that OWNERSHIP will be granted to
            tables = []

            for table in config["owns"]["tables"]:
                name_parts = table.split(".")
                info_schema = f"{name_parts[0].upper()}.INFORMATION_SCHEMA"

                if name_parts[2] == "*":
                    schemas = []
                    conn = SnowflakeConnector()

                    if name_parts[1] == "*":
                        db_schemas = conn.show_schemas(name_parts[0])

                        for schema in db_schemas:
                            if schema != info_schema:
                                schemas.append(schema)
                    else:
                        schemas = [f"{name_parts[0]}.{name_parts[1]}"]

                    for schema in schemas:
                        tables.extend(conn.show_tables(schema=schema))
                else:
                    tables = [table]

            # And then grant OWNERSHIP to all tables
            for db_table in tables:
                if self.check_grant_to_role(role, "OWNERSHIP", "TABLE", db_table):
                    already_granted = True
                else:
                    already_granted = False

                sql_commands.append(
                    {
                        "already_granted": already_granted,
                        "sql": GRANT_OWNERSHIP_TEMPLATE.format(
                            resource_type="TABLE",
                            resource_name=SnowflakeConnector.snowflaky(db_table),
                            role_name=SnowflakeConnector.snowflaky(role),
                        ),
                    }
                )
        except KeyError:
            logging.debug(
                "`owns.tables` not found for role {}, skipping generation of TABLE OWNERSHIP statements.".format(
                    role
                )
            )

        return sql_commands



import logging
from collections import namedtuple
from typing import Optional
from datetime import datetime, date

from .plugin.settings_service import PluginSettingsService, PluginSettingMissingError
from .project import Project
from .plugin import PluginType, PluginRef
from .db import project_engine
from .utils import nest, iso8601_datetime, coerce_datetime


class ScheduleAlreadyExistsError(Exception):
    """Occurs when a schedule already exists."""

    def __init__(self, schedule):
        self.schedule = schedule


Schedule = namedtuple(
    "Schedule",
    ("name", "extractor", "loader", "transform", "interval", "start_date", "env"),
)


class ScheduleService:
    def __init__(
        self, project: Project, plugin_settings_service: PluginSettingsService = None
    ):
        self.project = project
        self.plugin_settings_service = plugin_settings_service or PluginSettingsService(
            project
        )

    def add(
        self,
        session,
        name,
        extractor: str,
        loader: str,
        transform: str,
        interval: str,
        start_date: Optional[datetime] = None,
        **env,
    ):
        start_date = coerce_datetime(start_date) or self.default_start_date(
            session, extractor
        )
        schedule = Schedule(
            name, extractor, loader, transform, interval, start_date, env=env
        )

        return self.add_schedule(schedule)

    def default_start_date(self, session, extractor: str) -> datetime:
        """
        Returns the `start_date` of the extractor, or now.
        """
        extractor_ref = PluginRef(PluginType.EXTRACTORS, extractor)
        start_date = None
        try:
            start_date, _ = self.plugin_settings_service.get_value(
                session, extractor_ref, "start_date"
            )
        except PluginSettingMissingError:
            logging.debug(f"`start_date` not found in {extractor}.")

        # TODO: this coercion should be handled by the `kind` attribute
        # on the actual setting
        if isinstance(start_date, date):
            return coerce_datetime(start_date)

        if isinstance(start_date, datetime):
            return start_date

        return iso8601_datetime(start_date) or datetime.utcnow()

    def add_schedule(self, schedule: Schedule):
        with self.project.meltano_update() as meltano:
            # guard if it already exists
            if any(map(lambda s: s.name == schedule.name, self.schedules())):
                raise ScheduleAlreadyExistsError(schedule)

            # find the orchestrator plugin config
            schedules = nest(meltano, "schedules", value=[])
            schedules.append(self.schedule_definition(schedule))

        return schedule

    def schedules(self):
        return (
            self.yaml_schedule(schedule_def)
            for schedule_def in self.project.meltano.get("schedules", [])
        )

    @classmethod
    def schedule_definition(cls, schedule: Schedule) -> dict:
        definition = schedule._asdict()
        return dict(definition)

    @classmethod
    def yaml_schedule(cls, schedule_definition: dict) -> Schedule:
        return Schedule(**schedule_definition)



import logging

from typing import Dict, List, Tuple, Set

from meltano.core.permissions.utils.snowflake_connector import SnowflakeConnector

REVOKE_ROLE_TEMPLATE = "REVOKE ROLE {role_name} FROM {type} {entity_name}"

REVOKE_ALL_PRIVILEGES_ACCOUNTOBJECT_TEMPLATE = (
    "REVOKE {privileges} PRIVILEGES ON {resource_type} {resource_name} FROM ROLE {role}"
)

REVOKE_ALL_PRIVILEGES_OBJECT_TEMPLATE = "REVOKE {privileges} PRIVILEGES ON ALL {resource_type}S IN {parent_resource_type} {parent_resource_name} FROM ROLE {role}"


class SnowflakeRevokesGenerator:
    def __init__(self, grants_to_role: Dict, roles_granted_to_user: Dict) -> None:
        self.grants_to_role = grants_to_role
        self.roles_granted_to_user = roles_granted_to_user

    def generate_revoke_roles(
        self, entity_type: str, entity: str, config: str
    ) -> List[Dict]:
        """
        Generate the REVOKE statements for both roles and users.

        entity_type: "USER" or "ROLE"
        entity: the name of the entity (e.g. "yannis" or "REPORTER")
        config: the subtree for the entity as specified in the spec

        Returns the SQL commands generated as a list
        """
        sql_commands = []

        try:
            for member_role in config["member_of"]:
                granted_role = SnowflakeConnector.snowflaky(member_role).upper()

                sql_commands.append(
                    {
                        "already_granted": False,
                        "sql": REVOKE_ROLE_TEMPLATE.format(
                            role_name=SnowflakeConnector.snowflaky(member_role),
                            type=entity_type,
                            entity_name=SnowflakeConnector.snowflaky(entity),
                        ),
                    }
                )
        except KeyError:
            logging.debug(
                "`member_of` not found for {}, skipping generation of REVOKE ROLE statements.".format(
                    entity
                )
            )

        return sql_commands

    def generate_revoke_privileges_to_role(
        self, role: str, config: str, shared_dbs: Set
    ) -> List[Dict]:
        """
        Generate all the privilege revoking statements for a role.

        Most of the SQL command that will be generated are privileges revoked from
        roles and this function orchestrates the whole process

        role: the name of the role (e.g. "LOADER" or "REPORTER") the privileges
              are REVOKEd from
        config: the subtree for the role as specified in the spec
        shared_dbs: a set of all the shared databases defined in the spec.
                    Used down the road by generate_database_revokes() to also revoke
                    "IMPORTED PRIVILEGES" when access is for a shared DB.

        Returns the SQL commands generated as a list
        """
        sql_commands = []

        try:
            for warehouse in config["warehouses"]:
                new_commands = self.generate_warehouse_revokes(
                    role=role, warehouse=warehouse
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`warehouses` not found for role {}, skipping generation of Warehouse REVOKE statements.".format(
                    role
                )
            )

        try:
            for database in config["privileges"]["databases"]["read"]:
                new_commands = self.generate_database_revokes(
                    role=role, database=database, shared_dbs=shared_dbs
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.databases.read` not found for role {}, skipping generation of DATABASE read level REVOKE statements.".format(
                    role
                )
            )

        try:
            for database in config["privileges"]["databases"]["write"]:
                new_commands = self.generate_database_revokes(
                    role=role, database=database, shared_dbs=shared_dbs
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.databases.write` not found for role {}, skipping generation of DATABASE write level REVOKE statements.".format(
                    role
                )
            )

        try:
            for schema in config["privileges"]["schemas"]["read"]:
                database = schema.split(".")[0]
                new_commands = self.generate_database_revokes(
                    role=role, database=database, shared_dbs=shared_dbs
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.schemas.read` not found for role {}, skipping generation of DATABASE, via schema, read level REVOKE statements.".format(
                    role
                )
            )

        try:
            for schema in config["privileges"]["schemas"]["write"]:
                database = schema.split(".")[0]
                new_commands = self.generate_database_revokes(
                    role=role, database=database, shared_dbs=shared_dbs
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.schemas.write` not found for role {}, skipping generation of DATABASE, via schema, write level REVOKE statements.".format(
                    role
                )
            )

        try:
            for table in config["privileges"]["tables"]["read"]:
                new_commands = self.generate_table_and_view_revokes(
                    role=role, table=table, shared_dbs=shared_dbs
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.tables.read` not found for role {}, skipping generation of TABLE read level REVOKE statements.".format(
                    role
                )
            )

        try:
            for table in config["privileges"]["tables"]["write"]:
                new_commands = self.generate_table_and_view_revokes(
                    role=role, table=table, shared_dbs=shared_dbs
                )
                sql_commands.extend(new_commands)
        except KeyError:
            logging.debug(
                "`privileges.tables.write` not found for role {}, skipping generation of TABLE write level REVOKE statements.".format(
                    role
                )
            )

        return sql_commands

    def generate_warehouse_revokes(self, role: str, warehouse: str) -> List[str]:
        """
        Generate the REVOKE statements for Warehouse usage (only one type at the moment).

        role: the name of the role the privileges are REVOKEd to
        warehouse: the name of the warehouse (e.g. "transforming")

        Returns the SQL command generated
        """
        sql_commands = []

        sql_commands.append(
            {
                "already_granted": False,
                "sql": REVOKE_ALL_PRIVILEGES_ACCOUNTOBJECT_TEMPLATE.format(
                    privileges="ALL",
                    resource_type="WAREHOUSE",
                    resource_name=SnowflakeConnector.snowflaky(warehouse),
                    role=SnowflakeConnector.snowflaky(role),
                ),
            }
        )

        return sql_commands

    def generate_database_revokes(
        self, role: str, database: str, shared_dbs: Set
    ) -> List[str]:
        """
        Generate the REVOKE statements for Databases and Schemas.

        role: the name of the role the privileges are REVOKEd from
        database: the name of the database (e.g. "RAW")
        shared_dbs: a set of all the shared databases defined in the spec.

        Returns the SQL commands generated as a list
        """
        sql_commands = []
        already_granted = False

        # If this is a shared database, we have to revoke the "IMPORTED PRIVILEGES"
        #  privilege to the user and skip revoking the specific permissions as
        #  "Revoking individual privileges on imported databases is not allowed."
        if database in shared_dbs:
            sql_commands.append(
                {
                    "already_granted": already_granted,
                    "sql": REVOKE_ALL_PRIVILEGES_ACCOUNTOBJECT_TEMPLATE.format(
                        privileges="ALL IMPORTED",
                        resource_type="DATABASE",
                        resource_name=SnowflakeConnector.snowflaky(database),
                        role=SnowflakeConnector.snowflaky(role),
                    ),
                }
            )

            return sql_commands

        # And then revoke privileges to the database

        sql_commands.append(
            {
                "already_granted": already_granted,
                "sql": REVOKE_ALL_PRIVILEGES_ACCOUNTOBJECT_TEMPLATE.format(
                    privileges="ALL",
                    resource_type="DATABASE",
                    resource_name=SnowflakeConnector.snowflaky(database),
                    role=SnowflakeConnector.snowflaky(role),
                ),
            }
        )

        # Revoke privileges on all schemas in database

        sql_commands.append(
            {
                "already_granted": already_granted,
                "sql": REVOKE_ALL_PRIVILEGES_OBJECT_TEMPLATE.format(
                    privileges="ALL",
                    resource_type="SCHEMA",
                    parent_resource_type="DATABASE",
                    parent_resource_name=SnowflakeConnector.snowflaky(database),
                    role=SnowflakeConnector.snowflaky(role),
                ),
            }
        )
        return sql_commands

    def generate_table_and_view_revokes(
        self, role: str, table: str, shared_dbs: Set
    ) -> List[str]:
        """
        Generate the REVOKE statements for TABLEs and VIEWs.

        role: the name of the role the privileges are REVOKEd from
        table: the name of the TABLE/VIEW (e.g. "RAW.PUBLIC.MY_TABLE")
        shared_dbs: a set of all the shared databases defined in the spec.

        Returns the SQL commands generated
        """
        sql_commands = []

        # Split the table identifier into parts {DB_NAME}.{SCHEMA_NAME}.{TABLE_NAME}
        #  so that we can check and use each one
        name_parts = table.split(".")

        # Do nothing if this is a table inside a shared database:
        #  "Granting individual privileges on imported databases is not allowed."
        if name_parts[0] in shared_dbs:
            return sql_commands

        # Generate the INFORMATION_SCHEMA identifier for that database
        #  in order to be able to filter it out
        info_schema = f"{name_parts[0].upper()}.INFORMATION_SCHEMA"

        schemas = []

        conn = SnowflakeConnector()

        if name_parts[1] == "*":
            # If {DB_NAME}.*.* was provided as the identifier, we have to fetch
            #  each schema in database DB_NAME, so that we can revoke privileges
            #  for all tables in that schema
            # (You can not GRANT to all table with a wild card for the schema name)
            db_schemas = conn.show_schemas(name_parts[0])
            for schema in db_schemas:
                if schema != info_schema:
                    schemas.append(schema)
        elif "*" in name_parts[1]:
            # Similar to above, if {DB_NAME}.partial_*.* was provided,
            #  we need to find each of the schemas that match the pattern.
            conn = SnowflakeConnector()
            db_schemas = conn.show_schemas(name_parts[0])
            for db_schema in db_schemas:
                schema_name = db_schema.split(".", 1)[1].lower()
                if schema_name.startswith(name_parts[1].split("*", 1)[0]):
                    schemas.append(db_schema)
        else:
            schemas = [f"{name_parts[0]}.{name_parts[1]}"]

        for schema in schemas:
            # Revoke on ALL tables
            sql_commands.append(
                {
                    "already_granted": False,
                    "sql": REVOKE_ALL_PRIVILEGES_OBJECT_TEMPLATE.format(
                        privileges="ALL",
                        resource_type="TABLE",
                        parent_resource_type="SCHEMA",
                        parent_resource_name=SnowflakeConnector.snowflaky(schema),
                        role=SnowflakeConnector.snowflaky(role),
                    ),
                }
            )

            # Revoke on ALL views
            sql_commands.append(
                {
                    "already_granted": False,
                    "sql": REVOKE_ALL_PRIVILEGES_OBJECT_TEMPLATE.format(
                        privileges="ALL",
                        resource_type="VIEW",
                        parent_resource_type="SCHEMA",
                        parent_resource_name=SnowflakeConnector.snowflaky(schema),
                        role=SnowflakeConnector.snowflaky(role),
                    ),
                }
            )

        return sql_commands



from .cli import cli
from . import (
    elt,
    schema,
    discovery,
    initialize,
    add,
    install,
    invoke,
    ui,
    upgrade,
    permissions,
    schedule,
    select,
    repl,
    config,
    model,
)


def main():
    cli(obj={"project": None})



from itertools import groupby

from meltano.core.compiler.project_compiler import ProjectCompiler
from meltano.core.error import PluginInstallError
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginNotFoundError,
)
from meltano.core.plugin import PluginType
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.config_service import ConfigService
from meltano.core.plugin_install_service import PluginInstallService

from flask import Blueprint, request, jsonify, g

pluginsBP = Blueprint("plugins", __name__, url_prefix="/api/v1/plugins")


@pluginsBP.errorhandler(PluginInstallError)
def _handle(ex):
    return (jsonify({"error": True, "code": str(ex)}), 502)


@pluginsBP.route("/all", methods=["GET"])
def all():
    project = Project.find()
    discovery = PluginDiscoveryService(project)
    ordered_plugins = {}

    for type, plugins in groupby(discovery.plugins(), key=lambda p: p.type):
        ordered_plugins[type] = [plugin.canonical() for plugin in plugins]

    return jsonify(ordered_plugins)


@pluginsBP.route("/installed", methods=["GET"])
def installed():
    """Returns JSON of all installed plugins

    Fuses the discovery.yml data with meltano.yml data and sorts each type alphabetically by name
    """

    project = Project.find()
    config = ConfigService(project)
    discovery = PluginDiscoveryService(project)
    installed_plugins = {}

    # merge definitions
    for plugin in sorted(config.plugins(), key=lambda x: x.name):
        try:
            definition = discovery.find_plugin(plugin.type, plugin.name)
        except PluginNotFoundError:
            definition = {}

        merged_plugin_definition = {**definition.canonical(), **plugin.canonical()}
        merged_plugin_definition.pop("settings", None)
        merged_plugin_definition.pop("select", None)

        if not plugin.type in installed_plugins:
            installed_plugins[plugin.type] = []

        installed_plugins[plugin.type].append(merged_plugin_definition)

    return jsonify({**project.meltano, "plugins": installed_plugins})


@pluginsBP.route("/add", methods=["POST"])
def add():
    payload = request.get_json()
    plugin_type = PluginType(payload["plugin_type"])
    plugin_name = payload["name"]

    project = Project.find()
    add_service = ProjectAddService(project)
    plugin = add_service.add(plugin_type, plugin_name)

    return jsonify(plugin.canonical())


@pluginsBP.route("/install/batch", methods=["POST"])
def install_batch():
    payload = request.get_json()
    plugin_type = PluginType(payload["plugin_type"])
    plugin_name = payload["name"]

    project = Project.find()
    discovery = PluginDiscoveryService(project)
    target_plugin = discovery.find_plugin(plugin_type, plugin_name)

    config_service = ConfigService(project)
    add_service = ProjectAddService(project)
    install_service = PluginInstallService(project)
    ignored_types = [target_plugin.type, PluginType.TRANSFORMS]
    has_model = False
    batched = []
    for plugin in discovery.plugins():
        if plugin.namespace == target_plugin.namespace:
            if plugin.type not in ignored_types:
                add_service.add(plugin.type, plugin.name)
                plugin_install = config_service.find_plugin(
                    plugin.name, plugin_type=plugin.type
                )
                batched.append(plugin_install.canonical())
                run_venv = install_service.create_venv(plugin_install)
                run_install_plugin = install_service.install_plugin(plugin_install)
                if plugin.type is PluginType.MODELS:
                    has_model = True

    if has_model:
        compiler = ProjectCompiler(project)
        try:
            compiler.compile()
        except Exception as e:
            pass

    return jsonify(batched)


@pluginsBP.route("/install", methods=["POST"])
def install():
    payload = request.get_json()
    plugin_type = PluginType(payload["plugin_type"])
    plugin_name = payload["name"]

    project = Project.find()
    compiler = ProjectCompiler(project)
    install_service = PluginInstallService(project)
    config_service = ConfigService(project)

    plugin = config_service.find_plugin(plugin_name, plugin_type=plugin_type)
    run_venv = install_service.create_venv(plugin)
    run_install_plugin = install_service.install_plugin(plugin)

    if plugin_type is PluginType.MODELS:
        try:
            compiler.compile()
        except Exception as e:
            pass

    return jsonify(plugin.canonical())



import click
import datetime
import logging
import os
import sys

from . import cli
from .add import add_plugin, add_transform
from .params import project
from meltano.core.config_service import ConfigService
from meltano.core.runner.singer import SingerRunner
from meltano.core.runner.dbt import DbtRunner
from meltano.core.project import Project, ProjectNotFound
from meltano.core.job import Job
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginMissingError
from meltano.core.project_add_service import ProjectAddService
from meltano.core.transform_add_service import TransformAddService
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.db import project_engine
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginNotFoundError,
)
from meltano.core.logging import OutputLogger, JobLoggingService
from meltano.core.elt_context import ELTContextBuilder


@cli.command()
@click.argument("extractor")
@click.argument("loader")
@click.option("--dry", help="Do not actually run.", is_flag=True)
@click.option("--transform", type=click.Choice(["skip", "only", "run"]), default="skip")
@click.option(
    "--job_id", envvar="MELTANO_JOB_ID", help="A custom string to identify the job."
)
@project(migrate=True)
def elt(project, extractor, loader, dry, transform, job_id):
    """
    meltano elt EXTRACTOR_NAME LOADER_NAME

    extractor_name: Which extractor should be used in this extraction
    loader_name: Which loader should be used in this extraction
    """

    _, Session = project_engine(project)
    session = Session()
    job_logging_service = JobLoggingService(project)
    job = Job(
        job_id=job_id or f'job_{datetime.datetime.now().strftime("%Y%m%d-%H:%M:%S.%f")}'
    )

    # fmt: off
    with job.run(session), \
        job_logging_service.create_log(job.job_id, job.run_id) as log_file, \
        OutputLogger(log_file):

        try:
            install_missing_plugins(project, extractor, loader, transform)

            elt_context = (
                ELTContextBuilder(project)
                .with_job(job)
                .with_extractor(extractor)
                .with_loader(loader)
                .context(session)
            )

            if transform != "only":
                run_extract_load(elt_context, session, dry_run=dry)
            else:
                click.secho("Extract & load skipped.", fg="yellow")

            if transform != "skip":
                run_transform(elt_context, session, dry_run=dry, models=extractor)
            else:
                click.secho("Transformation skipped.", fg="yellow")
        except Exception as err:
            logging.error(f"ELT could not complete, an error happened during the process: {err}")
            raise click.Abort()
        finally:
            session.close()
    # fmt: on

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_elt(extractor=extractor, loader=loader, transform=transform)


def run_extract_load(elt_context, session, **kwargs):
    project = elt_context.project
    loader = elt_context.loader.ref
    extractor = elt_context.extractor.ref

    singer_runner = SingerRunner(
        elt_context,
        target_config_dir=project.meltano_dir(loader.type, loader.name),
        tap_config_dir=project.meltano_dir(extractor.type, extractor.name),
    )

    click.echo("Running extract & load...")
    singer_runner.run(session, **kwargs)
    click.secho("Extract & load complete!", fg="green")


def run_transform(elt_context, session, **kwargs):
    dbt_runner = DbtRunner(elt_context)
    click.echo("Running transformation...")
    dbt_runner.run(session, **kwargs)  # TODO: models from elt_context?
    click.secho("Transformation complete!", fg="green")


def install_missing_plugins(
    project: Project, extractor: str, loader: str, transform: str
):
    add_service = ProjectAddService(project)
    config_service = ConfigService(project)

    if transform != "only":
        try:
            config_service.find_plugin(extractor, plugin_type=PluginType.EXTRACTORS)
        except PluginMissingError:
            click.secho(
                f"Extractor '{extractor}' is missing, trying to install it...",
                fg="yellow",
            )
            add_plugin(add_service, project, PluginType.EXTRACTORS, extractor)

        try:
            config_service.find_plugin(loader, plugin_type=PluginType.LOADERS)
        except PluginMissingError:
            click.secho(
                f"Loader '{loader}' is missing, trying to install it...", fg="yellow"
            )
            add_plugin(add_service, project, PluginType.LOADERS, loader)

    if transform != "skip":
        try:
            config_service.find_plugin("dbt", plugin_type=PluginType.TRANSFORMERS)
        except PluginMissingError as e:
            click.secho(
                f"Transformer 'dbt' is missing, trying to install it...", fg="yellow"
            )
            add_plugin(add_service, project, PluginType.TRANSFORMERS, "dbt")

        transform_add_service = TransformAddService(project)
        try:
            # the extractor name should match the transform name
            plugin = config_service.find_plugin(
                extractor, plugin_type=PluginType.TRANSFORMS
            )

            # Update dbt_project.yml in case the vars values have changed in meltano.yml
            transform_add_service.update_dbt_project(plugin)
        except PluginMissingError:
            try:
                # Check if there is a default transform for this extractor
                PluginDiscoveryService(project).find_plugin(
                    PluginType.TRANSFORMS, extractor
                )

                click.secho(
                    f"Transform '{extractor}' is missing, trying to install it...",
                    fg="yellow",
                )
                add_transform(project, extractor)
            except PluginNotFoundError:
                # There is no default transform for this extractor..
                # Don't panic, everything is cool - just run custom transforms
                pass



import subprocess
import requests
import logging
from urllib.parse import urlsplit
from flask import (
    Blueprint,
    render_template,
    request,
    jsonify,
    redirect,
    g,
    current_app,
    send_from_directory,
)
from flask_security import login_required, roles_required
from jinja2 import TemplateNotFound

import meltano
from meltano.core.project import Project
from meltano.api.security import api_auth_required
from meltano.core.utils import truthy

logger = logging.getLogger(__name__)
root = Blueprint("root", __name__)


@root.errorhandler(500)
def internal_error(exception):
    logger.info(f"[{request.remote_addr}], error: {exception}")
    return jsonify({"error": str(exception)}), 500


@root.route("/-/dbt/", defaults={"path": "index.html"})
@root.route("/-/dbt/<path:path>")
def dbt_docs(path):
    project = Project.find()
    return send_from_directory(
        project.meltano_dir("transformers", "dbt", "target"), path
    )


# this route is a catch-all route to forward
# all oustanding request (not caught by any route)
# to the front-end.
@root.route("/", defaults={"path": ""})
@root.route("/<path:path>")
@login_required
def default(path):
    try:
        return render_template("webapp.html", jsContext=g.jsContext)
    except TemplateNotFound:
        return "Please run `make bundle` from src/webapp of the Meltano project."


@root.route("/upgrade", methods=["POST"])
@roles_required("admin")
def upgrade():
    meltano.api.executor.upgrade()
    return "Meltano update in progress.", 201


@root.route("/version")
def version():
    response_payload = {"version": meltano.__version__}

    if truthy(request.args.get("include_latest")):
        res = requests.get("https://pypi.org/pypi/meltano/json")
        pypi_payload = res.json()
        response_payload["latest_version"] = pypi_payload["info"]["version"]

    return jsonify(response_payload)


@root.route("/echo", methods=["POST"])
def echo():
    payload = request.get_json()
    print(payload)
    return jsonify(payload)



import os
import logging
import threading
import subprocess
import asyncio

from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler, EVENT_TYPE_MODIFIED
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.config_service import ConfigService, PluginMissingError
from meltano.core.plugin import PluginInstall, PluginType
from meltano.core.dbt_service import DbtService
from meltano.core.db import project_engine


class DbtEventHandler(PatternMatchingEventHandler):
    def __init__(self, queue):
        super().__init__(ignore_patterns=["*.log"])

        self._queue = queue

    def on_any_event(self, event):
        try:
            self._queue.put_nowait(event)
            self._queue._loop._write_to_self()

            logging.debug(f"Put {event} in the queue.")
        except asyncio.QueueFull:
            logging.debug(f"Discard {event} in the queue.")
        except Exception as e:
            logging.error(f"DbtWorker failed: {str(e)}")


class DbtWorker(threading.Thread):
    def __init__(self, project: Project, loader: str, loop=None):
        super().__init__()
        self.project = project
        self.loader = loader
        self.config_service = ConfigService(project)
        self.add_service = ProjectAddService(
            project, config_service=self.config_service
        )
        self.install_service = PluginInstallService(
            project, config_service=self.config_service
        )
        self.observer = None
        self._plugin = None
        self._loop = loop or asyncio.get_event_loop()

    @property
    def transform_dir(self):
        return self.project.root_dir("transform")

    def setup_observer(self, queue):
        # write every FS events in the Queue
        event_handler = DbtEventHandler(queue)

        observer = Observer()
        observer.schedule(event_handler, str(self.transform_dir), recursive=True)

        return observer

    async def process(self, session):
        dbt_service = DbtService(self.project)

        while True:
            # drain the queue
            while not self._queue.empty():
                self._queue.get_nowait()
                self._queue.task_done()

            # trigger the task
            try:
                loader = self.config_service.find_plugin(self.loader)
                await dbt_service.docs(session, loader, "generate")
            except PluginMissingError as err:
                logging.warning(
                    f"Could not generate dbt docs: '{str(err)}' is missing."
                )
            except:
                pass

            # wait for the next trigger
            logging.info("Awaiting task")
            await self._queue.get()
            self._queue.task_done()

            # wait for debounce
            await asyncio.sleep(5)

    def start(self):
        try:
            self._queue = asyncio.Queue(maxsize=1, loop=self._loop)

            self.observer = self.setup_observer(self._queue)
            self.observer.start()

            super().start()
        except OSError as err:
            # most probably INotify being full
            logging.warning(f"DbtWorker failed: INotify limit reached: {err}")

    def run(self):
        try:
            self._plugin = self.config_service.find_plugin("dbt")
        except PluginMissingError as err:
            self._plugin = self.add_service.add(PluginType.TRANSFORMERS, "dbt")
            self.install_service.install_plugin(self._plugin)

        _, Session = project_engine(self.project)

        try:
            session = Session()

            # TODO: this blocks the loop, we should probaly return a `Task` instance from
            # this function and let the caller schedule it on the loop
            # This class would not have to be a Thread an thus it could simplify the
            # handling of such cases in the future
            logging.info(
                f"Auto-generating dbt docs for in '{self.transform_dir}' for {self.loader}"
            )
            self._loop.run_until_complete(self.process(session))
        finally:
            session.close()

    def stop(self):
        if self.observer:
            self.observer.stop()



import os
import click
import logging
from pathlib import Path
from alembic.config import Config
from alembic.script import ScriptDirectory
from alembic.runtime.migration import MigrationContext
from alembic import command

from meltano.migrations import MIGRATION_DIR, LOCK_PATH, seed


class MigrationUneededException(Exception):
    """Occurs when no migrations are needed."""

    pass


class MigrationService:
    def __init__(self, engine):
        self.engine = engine

    def ensure_migration_needed(self, script, context, target_revision):
        current_head = context.get_current_revision()

        for rev in script.iterate_revisions(current_head, "base"):
            if rev.revision == target_revision:
                raise MigrationUneededException

    def upgrade(self):
        conn = self.engine.connect()
        cfg = Config()

        # this connection is used in `env.py` for the migrations
        cfg.attributes["connection"] = conn
        cfg.set_main_option("script_location", str(MIGRATION_DIR))
        script = ScriptDirectory.from_config(cfg)
        # let's make sure we actually need to migrate
        context = MigrationContext.configure(conn)

        try:
            # try to find the locked version
            HEAD = LOCK_PATH.open().read().strip()
            self.ensure_migration_needed(script, context, HEAD)

            click.secho(f"Upgrading database to {HEAD}")
            command.upgrade(cfg, HEAD)
        except FileNotFoundError:
            raise click.ClickException(
                "Cannot upgrade the system database, revision lock not found."
            )
        except MigrationUneededException:
            click.secho(f"System database up-to-date.")
        except Exception as err:
            click.secho(
                f"Cannot upgrade the system database. It might be corrupted or was created before database migrations where introduced (v0.34.0)",
                fg="yellow",
                err=True,
            )
            logging.exception(err)
        finally:
            conn.close()

    def seed(self, project):
        return seed(project)



"""Add run_id to Job

Revision ID: 53e97221d99f
Revises: 6ef30ab7b8e5
Create Date: 2019-10-10 13:12:55.147164

"""
from alembic import op
import sqlalchemy as sa
import sqlalchemy.orm
import sqlalchemy.types as types
from sqlalchemy import Column
from sqlalchemy.dialects.postgresql import UUID
import uuid


class GUID(types.TypeDecorator):
    """Platform-independent GUID type

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    Reference: https://docs.sqlalchemy.org/en/13/core/custom_types.html#backend-agnostic-guid-type
    """

    impl = types.CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(types.CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value


# revision identifiers, used by Alembic.
revision = "53e97221d99f"
down_revision = "6ef30ab7b8e5"
branch_labels = None
depends_on = None

Session = sa.orm.sessionmaker()


def upgrade():
    op.add_column("job", sa.Column("run_id", GUID))

    metadata = sa.MetaData(bind=op.get_bind())
    session = Session(bind=op.get_bind())
    Job = sa.Table("job", metadata, autoload=True)

    for job in session.query(Job):
        job.run_id = uuid.uuid4()

    session.commit()


def downgrade():
    op.drop_column("job", "run_id")



from typing import List

from meltano.core.permissions.pg_spec_loader import PGSpecLoader
from meltano.core.permissions.snowflake_spec_loader import SnowflakeSpecLoader
from meltano.core.permissions.utils.error import SpecLoadingError


def grant_permissions(
    db: str, spec_path: str, dry_run: bool, refresh: bool
) -> List[str]:
    if db == "postgres":
        if refresh:
            raise SpecLoadingError(f"Full refresh for {db} is not supported.")
        else:
            spec_loader = PGSpecLoader(spec_path)
    elif db == "snowflake":
        spec_loader = SnowflakeSpecLoader(spec_path)
    else:
        raise SpecLoadingError(f"Permissions Spec File for {db} is not supported.")

    sql_revoke_queries = []
    if refresh:
        sql_revoke_queries = spec_loader.generate_revoke_queries()

    sql_grant_queries = spec_loader.generate_permission_queries()

    sql_permission_queries = sql_revoke_queries + sql_grant_queries

    # just check at the moment
    # if not dry_run:
    #     # Do something

    # Always return the SQL commands that [would] run
    return sql_permission_queries



import logging
import subprocess
import asyncio
import os
import copy
from typing import Optional

from .project import Project
from .plugin import PluginInstall
from .plugin.error import PluginMissingError, PluginExecutionError
from .plugin.config_service import PluginConfigService
from .plugin.settings_service import PluginSettingsService
from .plugin_discovery_service import PluginDiscoveryService
from .venv_service import VenvService
from .error import Error, SubprocessError


def invoker_factory(project, plugin, *args, prepare_with_session=None, **kwargs):
    cls = PluginInvoker

    if hasattr(plugin.__class__, "__invoker_cls__"):
        cls = plugin.__class__.__invoker_cls__

    invoker = cls(project, plugin, *args, **kwargs)

    if prepare_with_session:
        invoker.prepare(prepare_with_session)

    return invoker


class InvokerNotPreparedError(Error):
    """Occurs when `invoke` is called before `prepare`"""

    pass


class PluginInvoker:
    """This class handles the invocation of a `Plugin` instance."""

    def __init__(
        self,
        project: Project,
        plugin: PluginInstall,
        plugin_config: Optional[dict] = None,
        run_dir=None,
        config_dir=None,
        venv_service: VenvService = None,
        plugin_config_service: PluginConfigService = None,
        plugin_settings_service: PluginSettingsService = None,
        plugin_discovery_service: PluginDiscoveryService = None,
    ):
        self.project = project
        self.plugin = plugin
        self._plugin_config = plugin_config
        self.venv_service = venv_service or VenvService(project)
        self.config_service = plugin_config_service or PluginConfigService(
            plugin,
            config_dir or self.project.plugin_dir(plugin),
            run_dir or self.project.run_dir(plugin.name),
        )
        self.settings_service = plugin_settings_service or PluginSettingsService(
            project
        )
        self.discovery_service = plugin_discovery_service or PluginDiscoveryService(
            project
        )
        self.plugin_def = self.discovery_service.find_plugin(
            self.plugin.type, self.plugin.name
        )
        self._prepared = False

    @property
    def capabilities(self):
        # we want to make sure the capabilites are immutable from the `PluginInvoker` interface
        return frozenset(self.plugin_def.capabilities)

    @property
    def select(self):
        return self.plugin.select or self.plugin_def.select or {"*.*"}

    @property
    def files(self):
        plugin_files = {**self.plugin.config_files, **self.plugin.output_files}

        return {
            _key: self.config_service.run_dir.joinpath(filename)
            for _key, filename in plugin_files.items()
        }

    @property
    def plugin_config(self):
        return self._plugin_config

    @plugin_config.setter
    def plugin_config(self, value):
        self._plugin_config = copy.deepcopy(value)
        # make sure to retrigger the 'configure' hook when
        # the plugin configuration has changed
        self._prepared = False

    def load_plugin_config(self, session):
        self._plugin_config = self._plugin_config or self.settings_service.as_config(
            session, self.plugin
        )

    def prepare(self, session):
        self.load_plugin_config(session)

        with self.plugin.trigger_hooks("configure", self):
            self.config_service.configure()
            self._prepared = True

    def exec_path(self):
        return self.venv_service.exec_path(
            self.plugin.executable,
            name=self.plugin.canonical_name,
            namespace=self.plugin.type,
        )

    def exec_args(self, *args):
        plugin_args = self.plugin.exec_args(self)

        return [str(arg) for arg in (self.exec_path(), *plugin_args, *args)]

    def Popen_options(self):
        return {}

    def invoke(self, *args, **Popen):
        if not self._prepared:
            raise InvokerNotPreparedError()

        Popen_options = self.Popen_options()
        Popen_options.update(Popen)

        process = None
        try:
            with self.plugin.trigger_hooks("invoke", self, args):
                popen_args = self.exec_args(*args)
                logging.debug(f"Invoking: {popen_args}")
                process = subprocess.Popen(popen_args, **Popen_options)
        except SubprocessError as perr:
            logging.error(f"{self.plugin.name} has failed: {str(perr)}")
            raise
        except Exception as err:
            logging.error(f"Failed to start plugin {self.plugin}.")
            raise

        return process

    async def invoke_async(self, *args, prepare=True, **Popen):
        if not self._prepared:
            raise InvokerNotPreparedError()

        Popen_options = self.Popen_options()
        Popen_options.update(Popen)

        process = None
        try:
            with self.plugin.trigger_hooks("invoke", self, args):
                process = await asyncio.create_subprocess_exec(
                    *self.exec_args(*args), **Popen_options
                )
        except SubprocessError as perr:
            logging.error(f"{self.plugin.name} has failed: {str(perr)}")
            raise
        except Exception as err:
            logging.error(f"Failed to start plugin {self.plugin}.")
            raise

        return process



import asyncio
import logging
import re
import sys


def setup_logging(log_level=logging.INFO):
    logging.basicConfig(
        stream=sys.stderr,
        format="[%(threadName)10s][%(levelname)s][%(asctime)s] %(message)s",
        level=log_level,
    )


def remove_ansi_escape_sequences(line):
    """
    Remove ANSI escape sequences that are used for adding colors in
     terminals, so that only the text is logged in a file
    """
    ansi_escape = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")
    return ansi_escape.sub("", line)


async def capture_subprocess_output(stream_captured, output_stream):
    """
    Capture in real time the output stream of a suprocess that is run async.
    The stream has been set to asyncio.subprocess.PIPE and is provided using
     stream_captured to this function.
    As new lines are captured for stream_captured, they are written to output_stream.
    This async function should be run with await asyncio.wait() while waiting
     for the subprocess to end.
    """
    while not stream_captured.at_eof():
        line = await stream_captured.readline()
        if not line:
            continue

        output_stream.write(line.decode())



import os
import yaml
import requests
import logging
import shutil
from copy import deepcopy
from typing import Dict, Iterator, Optional
from itertools import groupby, chain

import meltano.core.bundle as bundle
from .behavior.versioned import Versioned, IncompatibleVersionError
from .config_service import ConfigService
from .plugin import Plugin, PluginInstall, PluginType, PluginRef
from .plugin.factory import plugin_factory


class PluginNotFoundError(Exception):
    pass


class DiscoveryInvalidError(Exception):
    """Occurs when the discovery.yml fails to be parsed."""

    pass


class DiscoveryUnavailableError(Exception):
    """Occurs when the discovery.yml cannot be found or downloaded."""

    pass


MELTANO_DISCOVERY_URL = "https://www.meltano.com/discovery.yml"


class PluginDiscoveryService(Versioned):
    __version__ = 5

    def __init__(
        self,
        project,
        config_service: ConfigService = None,
        discovery: Optional[Dict] = None,
    ):
        self.project = project
        self.config_service = config_service or ConfigService(project)
        self._discovery = discovery

    @property
    def backend_version(self):
        return int(self._discovery.get("version", 1))

    @property
    def discovery(self):
        """
        Return first compatible discovery manifest from these locations:

        - project local `discovery.yml`
        - http://meltano.com/discovery.yml
        - .meltano/cache/discovery.yml
        - meltano.core.bundle
        """
        if self._discovery:
            return self._discovery

        try:
            local_discovery = self.project.root.joinpath("discovery.yml")

            if local_discovery.is_file():
                with local_discovery.open() as local:
                    self._discovery = yaml.load(local) or {}
            else:
                return self.fetch_discovery()

            self.ensure_compatible()

            return self._discovery
        except (yaml.YAMLError, FileNotFoundError) as e:
            raise DiscoveryInvalidError("discovery.yml is not well formed.") from e
        except IncompatibleVersionError as e:
            logging.fatal(
                "This version of Meltano cannot parse the plugins manifest, please update Meltano."
            )
            self._discovery = None
            raise

    def fetch_discovery(self):
        try:
            response = requests.get(MELTANO_DISCOVERY_URL)
            response.raise_for_status()
            self._discovery = yaml.load(response.text)
            self.ensure_compatible()

            with self.cached_discovery_file.open("w") as discovery_cache:
                discovery_cache.write(response.text)
            logging.debug(f"Discovery cache updated at {self.cached_discovery_file}")
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
            yaml.YAMLError,
        ) as e:
            # let's try loading the cache instead
            self._discovery = self.cached_discovery
        except IncompatibleVersionError as e:
            # let's default to the local cache
            if e.backend_version > e.version:
                logging.warning("You must update Meltano to get new plugins updates.")
            self._discovery = self.cached_discovery

        if not self._discovery:
            raise DiscoveryInvalidError("Cannot parse discovery.yml.")

        return self._discovery

    @property
    def cached_discovery(self):
        try:
            with self.cached_discovery_file.open() as discovery_cache:
                return yaml.load(discovery_cache)
        except (yaml.YAMLError, FileNotFoundError):
            logging.debug(
                f"Discovery cache corrupted or missing, falling back on the bundled discovery."
            )
            shutil.copy(bundle.find("discovery.yml"), self.cached_discovery_file)

            # load it back
            with self.cached_discovery_file.open() as discovery_cache:
                return yaml.load(discovery_cache)

    @property
    def cached_discovery_file(self):
        return self.project.meltano_dir("cache", "discovery.yml")

    def plugins(self) -> Iterator[Plugin]:
        custom_plugins = self.make_plugins(self.project.meltano.get("plugins", {}))
        discovery_plugins = self.make_plugins(self.discovery)

        return chain(custom_plugins, discovery_plugins)

    def make_plugins(self, plugin_defs: Dict) -> Iterator[Plugin]:
        plugins = deepcopy(plugin_defs)
        plugins.pop("version", 1)

        return (
            Plugin(
                plugin_type,
                plugin_def.pop("name"),
                plugin_def.pop("namespace"),
                **plugin_def,
            )
            for plugin_type, plugin_defs in plugins.items()
            for plugin_def in sorted(plugin_defs, key=lambda d: d["name"])
            if (PluginType.value_exists(plugin_type) and "namespace" in plugin_def)
        )

    def find_plugin(self, plugin_type: PluginType, plugin_name: str):
        name, _ = PluginRef.parse_name(plugin_name)
        try:
            return next(
                plugin
                for plugin in self.plugins()
                if (plugin.type == plugin_type and plugin.name == name)
            )
        except StopIteration:
            raise PluginNotFoundError(name)

    def discover(self, plugin_type: PluginType):
        """Return a pretty printed list of available plugins."""
        enabled_plugin_types = (
            (
                PluginType.EXTRACTORS,
                PluginType.LOADERS,
                PluginType.TRANSFORMERS,
                PluginType.MODELS,
                PluginType.TRANSFORMS,
                PluginType.ORCHESTRATORS,
            )
            if plugin_type == PluginType.ALL
            else (plugin_type,)
        )
        return {
            plugin_type: [p.name for p in plugins]
            for plugin_type, plugins in groupby(self.plugins(), lambda p: p.type)
            if plugin_type in enabled_plugin_types
        }

    def list_discovery(self, discovery):
        return "\n".join(
            plugin.name for plugin in self.plugins() if plugin.type == discovery
        )



import logging
import requests
import threading
import time
import webbrowser
from colorama import Fore


class UIAvailableWorker(threading.Thread):
    def __init__(self, url, open_browser=False):
        super().__init__()
        self._terminate = False

        self.url = url
        self.open_browser = open_browser

    def run(self):
        while not self._terminate:
            try:
                response = requests.get(self.url)
                if response.status_code == 200:
                    print(f"{Fore.GREEN}Meltano is available at {self.url}{Fore.RESET}")
                    if self.open_browser:
                        webbrowser.open(self.url)
                    self._terminate = True

            except:
                pass

            time.sleep(2)

    def stop(self):
        self._terminate = True



import contextlib
import io
import logging
from collections import OrderedDict
from typing import Iterator, Tuple


class hook:
    """
    This decorator marks a function as a __hook__.
    It will be found by the Hookable metaclass and
    registered to be triggered accordingly.
    """

    def __init__(self, hook_name, can_fail=False):
        self.name = hook_name
        self.can_fail = can_fail

    def __call__(self, func):
        func.__hook__ = self
        return func


class Hookable(type):
    """
    Metaclass that registers @hook functions into __hooks__
    Hooks are registered in declaration order.
    """

    def __new__(metacls, name, bases, dct):
        cls = type.__new__(metacls, name, bases, dct)
        cls.__hooks__ = {}

        for hook_name, hook in (
            (func.__hook__.name, func)
            for func in dct.values()
            if hasattr(func, "__hook__")
        ):
            cls.__hooks__[hook_name] = cls.__hooks__.get(hook_name, [])
            cls.__hooks__[hook_name].append(hook)

        return cls

    def __prepare__(name, bases, **kwds):
        return OrderedDict()


class HookObject(metaclass=Hookable):
    """
    Hook base class that handles the triggering of hooks.
    Hooks are triggered in reverse MRO order, which means
    derived classes hooks are called after their base class.
    """

    @contextlib.contextmanager
    def trigger_hooks(self, hook_name, *args, **kwargs):
        self.__class__.trigger(self, f"before_{hook_name}", *args, **kwargs)

        yield

        self.__class__.trigger(self, f"after_{hook_name}", *args, **kwargs)

    @classmethod
    def trigger(cls, target, hook_name, *args, **kwargs):
        hooks = [
            hook
            for hook_cls in reversed(cls.__mro__)
            if hasattr(hook_cls, "__hooks__")
            for hook in hook_cls.__hooks__.get(hook_name, [])
        ]

        for hook_func in hooks:
            try:
                hook_func(target, *args, **kwargs)
            except Exception as err:
                logging.warning(
                    f"{hook_name} hook '{hook_func.__name__}' has failed: {err}"
                )
                if not hook_func.__hook__.can_fail:
                    raise err



import os
import json
import yaml
import logging

from .project import Project
from .plugin import PluginType, PluginInstall, PluginRef
from .plugin_discovery_service import PluginDiscoveryService
from .plugin.factory import plugin_factory
from .config_service import ConfigService


class PluginNotSupportedException(Exception):
    pass


class PluginAlreadyAddedException(Exception):
    def __init__(self, plugin: PluginRef):
        self.plugin = plugin
        super().__init__()


class MissingPluginException(Exception):
    pass


class ProjectAddService:
    def __init__(
        self,
        project: Project,
        plugin_discovery_service: PluginDiscoveryService = None,
        config_service: ConfigService = None,
    ):
        self.project = project
        self.discovery_service = plugin_discovery_service or PluginDiscoveryService(
            project
        )
        self.config_service = config_service or ConfigService(project)

    def add(self, plugin_type: PluginType, plugin_name: str, **kwargs) -> PluginInstall:
        plugin = self.discovery_service.find_plugin(plugin_type, plugin_name)
        installed = plugin.as_installed()
        return self.config_service.add_to_file(installed)



import os
import json
import yaml
import logging
import click

from .project import Project
from .plugin import PluginType, PluginInstall
from .plugin.factory import plugin_factory
from .config_service import ConfigService


class ProjectAddCustomService:
    def __init__(self, project: Project, config_service: ConfigService = None):
        self.project = project
        self.config_service = config_service or ConfigService(project)

    def add(self, plugin_type: PluginType, plugin_name: str):
        namespace = click.prompt(
            "The `namespace` refers to the data source name.\n"
            "It is used to infer compatibility between components.\n\n"
            "(namespace)",
            type=str,
        )
        pip_url = click.prompt(
            "\nThe `pip URL` refers to the pip installation specification.\n"
            "You may provide any pip-compatible entry:\n"
            "\tfrom VCS: git+https://gitlab.com/meltano/tap-carbon-intensity.git\n"
            "\tfrom PyPI: tap-something\n"
            "\tfrom Source: [-e] /path/to/plugin/source\n\n"
            "(pip_url)",
            type=str,
        )
        executable = click.prompt(
            "\nThe `executable` refers to the entry point of the plugin.\n"
            "By convention, plugins should use their package name as executable, "
            "but it might be useful to override it to have multiple flavors of the same "
            "plugin installed.\n\n"
            "(executable)",
            default=plugin_name,
        )
        capabilities = click.prompt(
            "\nThe `capabilities` refer to the optional features the executable supports.\n"
            "Possible capabilities of extractors (taps) are:\n"
            "\t`catalog`: supports the `--catalog` flag\n"
            "\t`discover`: supports the `--discover` flag\n"
            "\t`properties`: supports the `--properties` flag\n"
            "\t`state`: supports the `--state` flag\n"
            "Multiple capabilities can be specified using a comma-separated string.\n\n"
            "(capabilities)",
            type=list,
            default=[],
            value_proc=lambda value: [c.strip() for c in value.split(",")],
        )

        # manually create the generic PluginInstall to save it
        # as a custom plugin
        install = PluginInstall(
            plugin_type,
            plugin_name,
            pip_url,
            capabilities=capabilities,
            namespace=namespace,
            executable=executable,
        )

        installed = self.config_service.add_to_file(install)
        click.secho(
            "The plugin has been added to your `meltano.yml`.\n"
            "If your plugin requires configuration options, you must add them directly "
            "in the created plugin definition under the `settings` section.\n\n"
            "Then, run `meltano install` to update the plugin configuration.",
            fg="yellow",
        )

        return installed



import os
import logging
import psutil
import meltano
import subprocess
import signal
import shutil
from typing import Optional

from meltano.core.project import Project
from meltano.core.migration_service import MigrationService
import meltano.core.bundle as bundle


class UpgradeError(Exception):
    """Occurs when the Meltano upgrade fails"""

    pass


class UpgradeService:
    def __init__(
        self, engine, project: Project, migration_service: MigrationService = None
    ):
        self.project = project
        self.migration_service = migration_service or MigrationService(engine)

    def reload(self):
        pid_file_path = self.project.run_dir("gunicorn.pid")
        try:
            with pid_file_path.open("r") as pid_file:
                pid = int(pid_file.read())

                process = psutil.Process(pid)
                process.send_signal(signal.SIGHUP)
        except Exception as ex:
            logging.error(f"Cannot restart from `{pid_file_path}`: {ex}")

    def upgrade_package(self, pip_url: Optional[str] = None, force=False):
        # we need to find out if the `meltano` module is installed as editable
        editable = meltano.__file__.endswith("src/meltano/__init__.py")
        editable = editable and not force

        if editable:
            logging.info(
                f"Skipping `meltano` upgrade because Meltano is installed as editable."
            )
        else:
            pip_url = pip_url or "meltano"
            run = subprocess.run(
                ["pip", "install", "--upgrade", pip_url],
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )

            if run.returncode != 0:
                raise UpgradeError(f"Failed to upgrade `meltano`.", run)

    def upgrade_files(self):
        """
        Update the files managed by Meltano inside the current project.
        """
        files_map = {
            bundle.find("dags/meltano.py"): self.project.root_dir(
                "orchestrate/dags/meltano.py"
            ),
            bundle.find("transform/profile/profiles.yml"): self.project.root_dir(
                "transform/profile/profiles.yml"
            ),
        }

        for src, dst in files_map.items():
            try:
                shutil.copy(src, dst)
                logging.info(f"{dst} has been updated.")
            except Exception as err:
                logging.error(f"Meltano could not update {dst}: {err}")

    def upgrade(self, **kwargs):
        self.upgrade_package(**kwargs)
        self.upgrade_files()
        self.migration_service.upgrade()
        self.migration_service.seed(self.project)



import logging

from .config_service import ConfigService
from .venv_service import VenvService
from .utils import noop
from .plugin import PluginInstall, Plugin, PluginRef
from .project import Project
from .error import (
    PluginInstallError,
    PluginInstallWarning,
    SubprocessError,
    PluginNotInstallable,
)


class PluginInstallService:
    def __init__(
        self,
        project: Project,
        venv_service: VenvService = None,
        config_service: ConfigService = None,
    ):
        self.project = project
        self.venv_service = venv_service or VenvService(project)
        self.config_service = config_service or ConfigService(project)

    def create_venv(self, plugin: PluginRef):
        return self.venv_service.create(namespace=plugin.type, name=plugin.name)

    def install_all_plugins(self, status_cb=noop):
        errors = []
        installed = []

        # TODO: config service returns PluginInstall, not Plugin
        for plugin in self.config_service.plugins():
            status = {"plugin": plugin, "status": "running"}
            status_cb(status)

            try:
                self.create_venv(plugin)
                self.install_plugin(plugin)

                status["status"] = "success"
                installed.append(status)
            except PluginInstallError as err:
                status["status"] = "error"
                status["message"] = str(err)
                status["details"] = err.process.stderr
                errors.append(status)
            except PluginInstallWarning as warn:
                status["status"] = "warning"
                status["message"] = str(warn)
                errors.append(status)
            except PluginNotInstallable as info:
                # let's totally ignore these plugins
                pass

            status_cb(status)

        return {"errors": errors, "installed": installed}

    def install_plugin(self, plugin: PluginInstall):
        if not plugin.installable():
            raise PluginNotInstallable()

        try:
            with plugin.trigger_hooks("install", self.project):
                self.create_venv(plugin)
                return self.venv_service.install(
                    namespace=plugin.type, name=plugin.name, pip_url=plugin.pip_url
                )
        except SubprocessError as err:
            raise PluginInstallError(
                f"{plugin.name} has an installation issue. {err}", err.process
            )



from .meltano_compiler_worker import MeltanoCompilerWorker
from .airflow_worker import AirflowWorker
from .dbt_worker import DbtWorker
from .ui_available_worker import UIAvailableWorker
from .api_worker import APIWorker



import os
import yaml
import logging
from typing import Dict, List, Optional

from meltano.core.utils import nest
from .project import Project
from .plugin import Plugin, PluginInstall, PluginType, PluginRef
from .plugin.factory import plugin_factory
from .plugin.error import PluginMissingError


class ConfigService:
    def __init__(self, project: Project):
        self.project = project

    def make_meltano_secret_dir(self):
        os.makedirs(self.project.meltano_dir(), exist_ok=True)

    def add_to_file(self, plugin: PluginInstall):
        installed_def = plugin.canonical()

        with self.project.meltano_update() as meltano_yml:
            if not plugin in self.plugins():
                plugins = nest(meltano_yml, f"plugins.{plugin.type}", value=[])
                plugins.append(installed_def)
            else:
                logging.warning(
                    f"{plugin.name} is already present, use `meltano install` to install it."
                )

        return plugin_factory(plugin.type, installed_def)

    def has_plugin(self, plugin_name: str):
        try:
            self.find_plugin(plugin_name)
            return True
        except PluginMissingError:
            return False

    def find_plugin(self, plugin_name: str, plugin_type: Optional[PluginType] = None):
        name, profile = PluginRef.parse_name(plugin_name)
        try:
            plugin = next(
                plugin
                for plugin in self.plugins()
                if (
                    plugin.name == name
                    and (plugin_type is None or plugin.type == plugin_type)
                )
            )

            plugin.profile = profile
            return plugin
        except StopIteration as stop:
            raise PluginMissingError(name) from stop

    def get_extractors(self):
        return filter(lambda p: p.type == PluginType.EXTRACTORS, self.plugins())

    def get_loaders(self):
        return filter(lambda p: p.type == PluginType.LOADERS, self.plugins())

    def get_transformers(self):
        return filter(lambda p: p.type == PluginType.TRANSFORMERS, self.plugins())

    def get_transforms(self):
        return filter(lambda p: p.type == PluginType.TRANSFORMS, self.plugins())

    def get_models(self):
        return filter(lambda p: p.type == PluginType.MODELS, self.plugins())

    def get_database(self, database_name):
        return yaml.load(
            open(self.project.meltano_dir(f".database_{database_name}.yml"))
        )

    def plugins(self) -> List[PluginInstall]:
        """Parse the meltano.yml file and return it as `PluginInstall` instances."""

        # this will parse the meltano.yml file and create an instance of the
        # corresponding `plugin_class` for all the plugins.
        return (
            plugin_factory(plugin_type, plugin_def)
            for plugin_type, plugin_defs in self.project.meltano.get(
                "plugins", {}
            ).items()
            for plugin_def in plugin_defs
        )



import os
import sqlalchemy
import logging
from typing import Iterable, Dict, Tuple, List
from copy import deepcopy
from enum import Enum

from meltano.core.utils import nest
from meltano.core.config_service import ConfigService
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.error import Error
from . import PluginRef, PluginType, Plugin, PluginInstall
from .error import PluginMissingError
from .setting import PluginSetting, REDACTED_VALUE


class PluginSettingMissingError(Error):
    """Occurs when a setting is missing."""

    def __init__(self, plugin: PluginRef, name: str):
        super().__init__(f"Cannot find setting {name} in {plugin}")


class PluginSettingValueSource(int, Enum):
    ENV = 0
    MELTANO_YML = 1
    DB = 2
    DEFAULT = 3


class SettingDefinition:
    def __init__(self, definition: dict):
        if "name" not in definition:
            raise ValueError("Setting definition must have a `name`.")

        self._definition = definition

    def __getattr__(self, attr):
        return getattr(self._definition, attr)

    def __getitem__(self, item):
        return self._definition[item]

    def __eq__(self, other):
        return self["name"] == other["name"]

    def __hash__(self):
        return hash(self["name"])


class PluginSettingsService:
    def __init__(
        self,
        project,
        config_service: ConfigService = None,
        plugin_discovery_service: PluginDiscoveryService = None,
    ):
        self.project = project
        self.config_service = config_service or ConfigService(project)
        self.discovery_service = plugin_discovery_service or PluginDiscoveryService(
            project
        )

    def as_config(
        self,
        session,
        plugin: PluginRef,
        sources: List[PluginSettingValueSource] = None,
        redacted=False,
    ) -> Dict:
        # defaults to the meltano.yml for extraneous settings
        plugin_install = self.get_install(plugin)
        plugin_def = self.get_definition(plugin)
        config = deepcopy(plugin_install.config)

        # definition settings
        for setting in self.definitions(plugin):
            value, source = self.get_value(session, plugin, setting["name"])
            if sources and source not in sources:
                continue

            # we don't want to leak secure informations
            # so we redact all `passwords`
            if redacted and value and setting.get("kind") == "password":
                value = REDACTED_VALUE

            nest(config, setting["name"], value)

        return config

    def as_env(
        self, session, plugin: PluginRef, sources: List[PluginSettingValueSource] = None
    ) -> Dict[str, str]:
        # defaults to the meltano.yml for extraneous settings
        plugin_def = self.get_definition(plugin)
        env = {}

        for setting in self.definitions(plugin):
            value, source = self.get_value(session, plugin, setting["name"])
            if sources and source not in sources:
                continue

            env_key = self.setting_env(setting, plugin_def)
            env[env_key] = str(value)

        return env

    def set(self, session, plugin: PluginRef, name: str, value, enabled=True):
        try:
            plugin_def = self.get_definition(plugin)
            setting_def = self.find_setting(plugin, name)
            env_key = self.setting_env(setting_def, plugin_def)

            if env_key in os.environ:
                logging.warning(f"Setting `{name}` is currently set via ${env_key}.")
                return

            if value == REDACTED_VALUE:
                return

            setting = PluginSetting(
                namespace=plugin.qualified_name, name=name, value=value, enabled=enabled
            )

            session.merge(setting)
            session.commit()

            return setting
        except StopIteration:
            logging.warning(f"Setting `{name}` not found.")

    def unset(self, session, plugin: PluginRef, name: str):
        session.query(PluginSetting).filter_by(
            namespace=plugin.qualified_name, name=name
        ).delete()
        session.commit()

    def get_definition(self, plugin: PluginRef) -> Plugin:
        return self.discovery_service.find_plugin(plugin.type, plugin.name)

    def find_setting(self, plugin: PluginRef, name: str) -> Dict:
        try:
            return next(
                setting
                for setting in self.definitions(plugin)
                if setting["name"] == name
            )
        except StopIteration:
            raise PluginSettingMissingError(plugin, name)

    def definitions(self, plugin_ref: PluginRef) -> Iterable[Dict]:
        settings = set()
        plugin_def = self.get_definition(plugin_ref)

        try:
            # there might be some settings declared on the PluginInstall
            # and such settings have precedence over the the definition
            plugin_install = self.get_install(plugin_ref)
            for setting_def in map(
                SettingDefinition, plugin_install._extras.get("settings", [])
            ):
                settings.add(setting_def)
        except PluginMissingError:
            pass

        # add the plugin definition settings
        for setting_def in map(SettingDefinition, plugin_def.settings):
            settings.add(setting_def)

        return list(settings)

    def get_install(self, plugin: PluginRef) -> PluginInstall:
        return self.config_service.find_plugin(plugin.name, plugin_type=plugin.type)

    def setting_env(self, setting_def, plugin_def):
        try:
            return setting_def["env"]
        except KeyError:
            parts = (plugin_def.namespace, setting_def["name"])
            process = lambda s: s.replace(".", "__").upper()
            return "_".join(map(process, parts))

    # TODO: ensure `kind` is handled
    def get_value(self, session, plugin: PluginRef, name: str):
        plugin_install = self.get_install(plugin)
        plugin_def = self.get_definition(plugin)
        setting_def = self.find_setting(plugin, name)

        try:
            env_key = self.setting_env(setting_def, plugin_def)

            # priority 1: environment variable
            if env_key in os.environ:
                logging.debug(
                    f"Found ENV variable {env_key} for {plugin_def.namespace}:{name}"
                )
                return (os.environ[env_key], PluginSettingValueSource.ENV)

            # priority 2: installed configuration
            if setting_def["name"] in plugin_install.config:
                return (
                    plugin_install.config[setting_def["name"]],
                    PluginSettingValueSource.MELTANO_YML,
                )

            # priority 3: settings database
            return (
                (
                    session.query(PluginSetting)
                    .filter_by(namespace=plugin.qualified_name, name=name, enabled=True)
                    .one()
                    .value
                ),
                PluginSettingValueSource.DB,
            )
        except sqlalchemy.orm.exc.NoResultFound:
            # priority 4: setting default value
            # that means it was not overriden
            return setting_def.get("value"), PluginSettingValueSource.DEFAULT



import logging
import threading
import time
import psutil

from meltano.core.project import Project
from meltano.core.plugin import PluginInstall, PluginType
from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.config_service import ConfigService, PluginMissingError
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.db import project_engine
from meltano.core.utils.pidfile import PIDFile, UnknownProcessError


class AirflowWorker(threading.Thread):
    def __init__(self, project: Project):
        super().__init__(name="AirflowWorker")

        self.project = project
        self.config_service = ConfigService(project)
        self.add_service = ProjectAddService(
            project, config_service=self.config_service
        )
        self.install_service = PluginInstallService(
            project, config_service=self.config_service
        )
        self._plugin = None
        self._webserver = None
        self._scheduler = None

    def kill_stale_workers(self):
        stale_workers = []
        workers_pid_files = map(self.pid_file, ("webserver", "scheduler"))

        for pid_file in workers_pid_files:
            try:
                stale_workers.append(pid_file.process)
            except UnknownProcessError:
                pass

        def on_terminate(process):
            logging.info(f"Process {process} ended with exit code {process.returncode}")

        for process in stale_workers:
            logging.debug(f"Process {process} is stale, terminating it.")
            process.terminate()

        gone, alive = psutil.wait_procs(stale_workers, timeout=5, callback=on_terminate)

        # kill the rest
        for process in alive:
            process.kill()

        for pid_file in workers_pid_files:
            try:
                pid_file.unlink()
            except:
                pass

    def start_all(self):
        _, Session = project_engine(self.project)
        logs_dir = self.project.run_dir("airflow", "logs")

        try:
            session = Session()
            invoker = invoker_factory(
                self.project, self._plugin, prepare_with_session=session
            )

            # fmt: off
            with logs_dir.joinpath("webserver.log").open("w") as webserver, \
              logs_dir.joinpath("scheduler.log").open("w") as scheduler:
                self._webserver = invoker.invoke("webserver", "-w", "1", stdout=webserver)
                self._scheduler = invoker.invoke("scheduler", stdout=scheduler)

                self.pid_file("webserver").write_pid(self._webserver.pid)
                self.pid_file("scheduler").write_pid(self._scheduler.pid)
            # fmt: on

            # Time padding for server initialization so UI iframe displays as expected
            # (iteration potential on approach but following UIAvailableWorker sleep approach)
            time.sleep(2)
        finally:
            session.close()

    def pid_file(self, name) -> PIDFile:
        return PIDFile(self.project.run_dir("airflow", f"{name}.pid"))

    def run(self):
        try:
            self._plugin = self.config_service.find_plugin("airflow")
        except PluginMissingError as err:
            self._plugin = self.add_service.add(PluginType.ORCHESTRATORS, "airflow")
            self.install_service.install_plugin(self._plugin)

        self.kill_stale_workers()
        self.start_all()

    def stop(self):
        self.kill_stale_workers()



from flask_security import UserMixin, RoleMixin

from . import db


class User(db.Model, UserMixin):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(255), unique=True, index=True)
    email = db.Column(db.String(255), unique=True)
    password = db.Column(db.String(255))
    active = db.Column(db.Boolean())
    confirmed_at = db.Column(db.DateTime())
    roles = db.relationship(
        "Role", secondary="roles_users", backref=db.backref("users", lazy="dynamic")
    )

    def __repr__(self):
        roles = ",".join(r.name for r in self.roles)
        return f"<User username={self.username} roles={roles}>"

    def __str__(self):
        roles = ",".join(r.name for r in self.roles)
        return f"@{self.username}({roles})"

    def canonical(self, _scopes):
        return self.username


class RolesUsers(db.Model):
    __tablename__ = "roles_users"
    id = db.Column(db.Integer(), primary_key=True)
    user_id = db.Column("user_id", db.Integer(), db.ForeignKey("user.id"))
    role_id = db.Column("role_id", db.Integer(), db.ForeignKey("role.id"))


class RolePermissions(db.Model):
    __tablename__ = "role_permissions"
    id = db.Column(db.Integer(), primary_key=True)
    role_id = db.Column("role_id", db.Integer(), db.ForeignKey("role.id"))
    type = db.Column(db.String())
    context = db.Column(db.String())
    role = db.relationship("Role", backref="permissions")

    def canonical(self, scopes=[]):
        canonical = {
            "role_id": self.role_id,
            "type": self.type,
            "context": self.context,
        }

        if Role in scopes:
            del canonical["role_id"]

        return canonical


class Role(db.Model, RoleMixin):
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.String(80), unique=True)
    description = db.Column(db.String(255))

    def __eq__(self, other):
        if isinstance(other, Role):
            return super().__eq__(other)

        return self.name == other

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return f"<Role({self.name})>"

    def canonical(self, _scopes):
        return self.name
